/*
  ambiore program from the AMBIORE package.

  (C) Copyright 2004, Dick Hwang and University of Washington.

  dhwang@u.washington.edu
  http://www.phrap.org

 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/time.h>
#include "utility.h"

#define USAGE "USAGE: ambiore [options] seqfile treefile rate_categories_file root_categories_file\n"\
              "  -burnin #     Number of iterations of updating all missing data before collecting samples (default: 40)\n" \
              "  -samples #    Number of samples to collect (default: 50,000)\n"\
              "  -rate #.#     Rate of sample collection per position-specific missing data update (default 0.01)\n"\
              "  -update_parameters_rate #.#  Probability of updating branch scale and mutation rate parameters between each position-specific missing data update (default 0.002)\n"\
              "  -update_root_rate rate #.#   Probability of updating root and inserted sequence transition parameters between each position-specific missing data update (default 0.05)\n"\
              "  -map file     File indicating mapping of position coordinates in category files with seqfile\n"\
              "  -seed #       Integer to seed random number generator\n"\
              "  -out prefix   Pathname prefix for output files\n"\
              "  -node_count   Output position-specific counts of root nucleotides in sample\n"\
              "  -mutation_types # [file]    Use # mutation types as specified in file (file omitted if 1 mutation type)\n"\
              "\n"\
              "(C) Copyright 2004, Dick Hwang and the University of Washington\n"\
              "                    www.phrap.org\n"

#define DEFAULT_SAMPLES 50000
#define DEFAULT_SAMPLE_RATE 0.01
#define DEFAULT_BURNIN 40
#define DEFAULT_ALPHA 2.0
#define DEFAULT_MAX_MUTATION_RATE 500000
#define DEFAULT_UPDATE_PARAMETERS_RATE 0.002
#define DEFAULT_UPDATE_ROOT_RATE 0.05

#define PREBURNIN 3

#define NUM_ROOT_PARAMETERS 48
#define NUM_TOTAL_ROOT_PARAMETERS 64

#define NUM_ROOT_BASE 16

#define MAX_CHILDREN 16
#define MAX_BRANCHES 256

#define NUM_NUC 4
#define GAP 5

#define MAX_NAME 128
#define MAX_LINE 2048

#define NODE (-1)
#define AXIL (-2)
#define BRANCH (-3)

#define T 0
#define C 1
#define A 2
#define G 3

#define FLOAT double

typedef struct NodeS {
  struct NodeS *children[MAX_CHILDREN];
  struct NodeS *parent;
  int n_children;
  int time_to_parent;
  int seq_num;
  int configuration_index;
  int num;
} Node;

typedef struct {
  unsigned short el[NUM_NUC];
} NodeCount;

typedef struct {
  char el[NUM_NUC];
} NucVec;

typedef struct {
  short el[NUM_NUC];
} ShortNucVec;

typedef struct {
  FLOAT el[NUM_NUC][NUM_NUC][NUM_NUC][NUM_NUC];
} ParamTable;

typedef struct {
  FLOAT el[NUM_NUC][NUM_NUC][NUM_NUC];
} RootTable;

typedef struct {
  char* seq;
  char name[MAX_NAME];
} Sequence;

typedef struct {
  Node* node;
  signed char type;
} Map;

typedef struct {
  char* configuration;
  char* old_config;
  int pos;
} ConfigP;

typedef struct {
  double el[NUM_ROOT_BASE][NUM_NUC-1][NUM_NUC-1];
  double inverse[NUM_ROOT_BASE][NUM_NUC];
} RootVariance;

typedef struct {
  int right; /* 3' neighbor of branch excluding end node */
  int left;
  int node_right; /* 3' neighbor of end node */
  int node_left;
} Neighbors;

typedef struct {
  Neighbors* neighbors;
  unsigned char rate_category;
  char present; /* 0 = absent, 1 = present, 2 = end node is present */
  signed char direction; /* -1 = reverse, 0 = symmetric, 1 = forward */
} BranchState;

typedef struct pc {
  unsigned char change;
  unsigned char num;
} PositionChanges;

typedef struct {
  int n_seq;
  int num_branches;
  int num_parameters[2];
  int num_mutation_parameters[2];
  int num_total_mutation_parameters[2];
  int seq_length;
  int tree_time;
  int samples;
  int sample_rate;
  int burnin;

  int acceptance_numer;
  int acceptance_denom;
  int mutation_acceptance_numer;
  int mutation_acceptance_denom;
  int branch_acceptance_numer;
  int branch_acceptance_denom;
  int root_acceptance_numer;
  int root_acceptance_denom;
  int rate_multiplier_acceptance_numer;
  int rate_multiplier_acceptance_denom;

  Sequence* sequence;
  int* map;
  int map_length;

  BranchState **branch_state; /* indexed by sequence position, then branch */
  short* root_position; /* location of root at each sequence position */
  int num_rate_categories;
  char* category_type; /* 0 = symmetric, 1 = asymmetric */
  char** category_name; 
  char** category_branch_existence; /* indexed by category, then branch; indicates whether a rate category exists for a particular branch */
  char* root_category;
  int num_root_categories;
  char** root_category_name;

  char* out_prefix;
  FILE** branch_out; /* indexed by mutation type */
  FILE** mutation_out;
  FILE** root_out;

  char** configurations;
  Map* configuration_map;
  Node* root;
  FLOAT param_table[NUM_NUC][NUM_NUC][NUM_NUC][NUM_NUC]; 
  ParamTable*** branch_param_table; /* indexed by hidden state, then rate category, then branch num */
  ParamTable*** reverse_branch_param_table;
  double**** branch_param_list; /* indexed by hidden state, then rate category, then branch num, then parameter */
  RootTable* cond_root; /* indexed by root category */
  RootTable* reverse_cond_root;
  int seed;
  int record;
  int metropolis;
  int probability;
  double max_rate;
  double alpha;
  double update_parameters_rate;
  double update_root_rate;

  int** root_changes;
  int**** changes; /* indexed by hidden state, then category, then branch, then parameter */

  FLOAT** parameters;
  FLOAT** root_transitions;

  int num_time_categories; /* number of time categories partitioning the branches */
  int* time_category; /* time category of each branch */
  int* rate_time_category; /* time category of each rate category */

  double* branch_ratio[MAX_BRANCHES]; /* indexed by branch, then mutation type */
  int branch_length[MAX_BRANCHES];
  int num_mutation_types;
  int mutation_types;
  int* mutation_type[2]; /* indexed by symmetry, then mutation parameter */
  int* mutation_type_symmetry; /* indexed by mutation type; 0 = symmetric, 1 = asymmetric, 2 = both */
  int* num_mutation_parameters_per_type[2]; /* indexed by symmetry, then mutation type */
  int** mutation_parameters_per_type[2]; /* indexed by symmetry, then mutation type, then mutation parameter for that type */
  int* num_mutation_types_per_total_parameters[2]; /* indexed by symmetry, then total parameter */
  int** mutation_types_per_total_parameters[2]; /* indexed by symmetry, then total parameter, then mutation type */
  int* num_total_parameters_per_type[2];
  int** total_parameters_per_type[2];

  int trinuc_parameters[2][256]; 
  int complement_trinuc_parameters[256];
  int* param_complement;
  int*** param_enum[2];
  int** param_enum_sum[2];
  int* param_enum_addend[2];
  int num_mutation_set[2];
  int root_parameters[NUM_NUC][NUM_NUC][NUM_NUC];
  int complement_root_parameters[NUM_NUC][NUM_NUC][NUM_NUC];
  int root_parameters_vector[64];
  int complement_root_parameters_vector[64];
  int root_enum[NUM_TOTAL_ROOT_PARAMETERS][3];
  int root_enum_sum[NUM_ROOT_BASE][NUM_NUC];
  int root_enum_addend[NUM_ROOT_PARAMETERS];


  NodeCount* node_count; /* indexed by sequence position, then internal node */
  FILE* node_out;
  FILE*** changes_out; /* indexed by category, then branch */
  FILE* variation_out;
  FILE* transition_out;
  FILE* rate_multiplier_out;
  FILE* initial_out;

  double trinuc_freqs[2][NUM_NUC][NUM_NUC][NUM_NUC];

  int variation;
  int num_hidden_states;
  int fix_transition_parameters;
  unsigned char* hidden_state; /* indexed by sequence position */
  double **transition; /* indexed by hidden state, then hidden state */
  int** fix_transition; /* indexed by hidden state, then hidden state */
  int* last_unfixed_transition; /* indexed by hidden state */
  double *initial; /* indexed by hidden state */
  double **emission; /* indexed by sequence position, then hidden state */
  /*  double **forward; */ /* indexed by sequence position, then hidden state */
  double **backward; /* indexed by sequence position, then hidden state */
  double *rate_multiplier; /* indexed by hidden state */
  int *fix_rate_multiplier; /* indexed by hidden state */
  unsigned short** variation_count; /* indexed by sequence position, then hidden state */

  int stats;
} Params;

char NAUC[7] = "TCAGN-";
int complement[4] = {2,3,0,1};

void initiate_configuration(Params* params);
int get_state(int n, double* vector);
void create_configuration(Params* params);
void allocate_configuration(Params* params);
int configuration_index(Params* params);
void node_index(Params* params, Node* node, int* index);
void read_params(int argc, char* argv[], Params* params);
void read_parameters(Params* params, FLOAT* parameters, int a, char* filename);
void read_root_parameters(Params* params, FLOAT* parameters, char* filename);
void read_tree(Params* params, char* treefile);
Node* read_node(Params* params, FILE* in, Node* parent);
void read_sequence(Params* params, char* seqfile);
void fill_configuration_map(Params* params, Node* node);
void mcmc(Params* params);
void print_changes(Params* params);
void count_configuration_changes(Params* params);
void count_changes(Params* params, ConfigP* configp);
int get_tree_state(int tree_size, double* tree);
void newize_configuration(Params* params, int i, char* configuration, int new);
void conditional_likelihood(Params* params, ConfigP* configp, FLOAT** likelihood);
int gibbs_sample(Params* params, ConfigP* configp, FLOAT** likelihood);
void iterate_gibbs(Params* params, int pos, char* configuration, FLOAT** likelihood);
void count_changes_(Params* params, int earliest, ConfigP* configp);
void print_ratios(Params* params);
int iterate_mutation_parameters(Params* params);
void sample_mutation_parameters(Params* params, int c, double* parameters, double* mle, double* variance);
void find_mutation_mle(Params* params, int c, double* parameters, double* variance);
void make_trinuc_parameters(Params* params, char* mutation_type_file);
void make_root_parameters(Params* params);
void print_parameters(Params* params);
int iterate_branch_parameters(Params* params);
void sample_branch_parameters(Params* params, double** parameters, double** mle, double** variance);
void find_branch_mle(Params* params, double** parameters, double** variance);
int iterate_root_parameters(Params* params);
void update_root_table(Params* params);
void sample_root_parameters(Params* params, int c, double* parameters, double* mle, RootVariance* variance);
void find_root_mle(Params* params, int c, double* parameters, RootVariance* variance);
void update_branch_param_table(Params* params);
void enstate_branches(Params* params);
void instate_node1(Params* params, Node* node);
void instate_node2(Params* params, Node* node);
void instate_node3(Params* params, Node* node);
void read_rate_categories(Params* params, char* filename);
void read_root_categories(Params* params, char* filename);
void read_mapfile(Params* params, char* mapfile);
void normalize_rates(Params* params);
void print_categories(Params* params);
double calculate_probability(Params* params);
void count_nodes(Params* params);
void print_nodes(Params* params);
void markov(Params* params);
void emission_probabilities(Params* params);
void forward_variables(Params* params);
void backward_variables(Params* params);
void sample_hidden_states(Params* params);
void count_position_changes(Params* params, PositionChanges** position_changes, int pos);
void initiate_variation(Params* params);
void allocate_variation(Params* params);
void iterate_variation(Params* params);
void iterate_variation_parameters(Params* params);
void print_variation_parameters(Params* params);
void update_branch_param_list(Params* params);
void sample_from_counts(int n, int* count, double* parameters, int* fixed);
void find_rate_multiplier_mle(Params* params, double* parameters, double* variance);
void sample_rate_multiplier_parameters(Params* params, double* parameters, double* mle, double* variance);
int iterate_rate_multiplier_parameters(Params* params);
void iterate_parameters(Params* params);
void compute_stationary_from_transition(Params* params);
void matrix_multiply(double** c, double** a, double** b, int n);

int main(int argc, char* argv[]) {
  Params *params;
  char randomstate[256];

  (params = (Params*)malloc(sizeof(Params))) || die("Could not allocate memory for params\n");
  initstate(1, randomstate, 256);
  read_params(argc, argv, params);
  fprintf(stderr, "Initiating configuration...\n");
  initiate_configuration(params);
  initiate_variation(params);

  fprintf(stderr, "Beginning MCMC iterations...");
  fprintf(stderr, "tree size = %d\n", params->tree_time);
  mcmc(params);

  return 0;
}

void mcmc(Params* params) {
  char* configuration;
  int i, j, k, h;
  struct timeval tv1, tv2;
  double tm;
  FLOAT** likelihood;
  int* permutation;
  int tmp, r;
  double time_spent, total_time;
  double percent_complete;
  double total_iterations;

  (configuration = (char*)malloc(sizeof(char) * params->tree_time)) || die("Could not allocate memory for configuration\n");
  (likelihood = (FLOAT**)malloc(sizeof(FLOAT*) * params->tree_time)) || die("Could not allocate memory for likelihood\n");
  for(i = 0; i < params->tree_time; i++)
    (likelihood[i] = (FLOAT*)malloc(sizeof(FLOAT) * NUM_NUC)) || die("Could not allocate memory for likelihood[%d]\n");

  (params->changes = (int****)malloc(sizeof(int***) * params->num_hidden_states)) || die("Could not allocate memory for params->changes\n");
  for(h = 0; h < params->num_hidden_states; h++) {
    (params->changes[h] = (int***)malloc(sizeof(int**) * params->num_rate_categories)) || die("Could not allocate memory for params->changes[%d]\n", h);
    for(i = 0; i < params->num_rate_categories; i++) {
      (params->changes[h][i] = (int**)malloc(sizeof(int*) * params->num_branches)) || die("Could not allocate memory for params->changes[%d][%d]\n", h, i);
      for(j = 0; j < params->num_branches; j++)
	(params->changes[h][i][j] = (int*)malloc(sizeof(int) * params->num_total_mutation_parameters[params->category_type[i]])) || die("Could not allocate memory for params->changes[%d][%d][%d]\n", h, i, j);
    }
  }
  if(params->node_out) {
    (params->node_count = (NodeCount*)malloc(sizeof(NodeCount) * params->seq_length)) || die("Could not allocate memory for params->node_count\n");
    for(i = 0; i < params->seq_length; i++)
      for(j = 0; j < NUM_NUC; j++)
	params->node_count[i].el[j] = 0;
  }
  (params->root_changes = (int**)malloc(sizeof(int*) * params->num_root_categories)) || die("Could not allocate memory for params->root_changes\n");
  for(i = 0; i < params->num_root_categories; i++)
    (params->root_changes[i] = (int*)malloc(sizeof(int) * NUM_TOTAL_ROOT_PARAMETERS)) || die("Could not allocate memory for params->root_changes[%d]\n", i);

  (permutation = (int*)malloc(sizeof(int) * params->seq_length)) || die("Could not allocte memory for mutations\n");

  for(i = 0; i < params->seq_length; i++)
    permutation[i] = i;

  time_spent = 0.0;
  percent_complete = 0.0;
  total_iterations = (PREBURNIN / 1.3 + params->burnin) * params->seq_length + params->samples * params->sample_rate;

  fprintf(stderr, "Burning in configuration...\n");
  params->record = 0;
  for(i = 0; i < PREBURNIN; i++) {
    if(params->stats)
      fprintf(stderr, "  Iteration %d...", i);
    gettimeofday(&tv1, NULL);
    for(j = 0; j < params->seq_length; j++) {
      r = random_int(j, params->seq_length - 1);
      tmp = permutation[j];
      permutation[j] = permutation[r];
      permutation[r] = tmp;
    }
    for(j = 0; j < params->seq_length; j++) {
      if(params->root_position[permutation[j]] != -1)
	 iterate_gibbs(params, permutation[j], configuration, likelihood);
    }

    gettimeofday(&tv2, NULL);
    tm = tv2.tv_sec - tv1.tv_sec + (tv2.tv_usec - tv1.tv_usec) / 1000000.0;
    time_spent += tm;
    percent_complete += params->seq_length / 1.3 / total_iterations;
    total_time = time_spent / percent_complete;
    if(params->stats)
      fprintf(stderr, " %g seconds\n", tm);
    fprintf(stderr, "%2.3f%% complete, %.2d:%2.2d:%2.2ds elapsed (~%.2d:%2.2d:%2.2ds total)\n", percent_complete * 100.0, (int)time_spent / 3600, (int)time_spent % 3600 / 60, (int)time_spent % 60, (int)total_time / 3600, (int)total_time % 3600 / 60, (int)total_time % 60);
  }

  fprintf(stderr, "Burning in parameters...\n");
  count_configuration_changes(params);
  params->record = 1;

  for(i = 0; i < params->burnin; i++) {
    normalize_rates(params);
    if(0 && i < params->burnin / 3)
      params->metropolis = 0;
    else
      params->metropolis = 1;

    if(params->stats)
      fprintf(stderr, "  Iteration %d...", i);
    gettimeofday(&tv1, NULL);
    for(j = 0; j < params->seq_length; j++) {
      r = random_int(j, params->seq_length - 1);
      tmp = permutation[j];
      permutation[j] = permutation[r];
      permutation[r] = tmp;
    }
    for(j = 0; j < params->seq_length; j++) {
      if(params->root_position[permutation[j]] != -1) {
	iterate_gibbs(params, permutation[j], configuration, likelihood);
	if(random_double(0.0, 1.0) <= params->update_parameters_rate)
	  iterate_parameters(params);
	if(random_double(0.0, 1.0) <= params->update_root_rate && iterate_root_parameters(params))
	  update_root_table(params);
      }
    }
    if(params->variation)
      iterate_variation(params);

    gettimeofday(&tv2, NULL);
    tm = tv2.tv_sec - tv1.tv_sec + (tv2.tv_usec - tv1.tv_usec) / 1000000.0;
    time_spent += tm;
    percent_complete += params->seq_length / total_iterations;
    total_time = time_spent / percent_complete;
    if(params->stats)
      fprintf(stderr, " %g seconds\n", tm);
    fprintf(stderr, "%2.3f%% complete, %.2d:%2.2d:%2.2ds elapsed (~%.2d:%2.2d:%2.2ds total)\n", percent_complete * 100.0, (int)time_spent / 3600, (int)time_spent % 3600 / 60, (int)time_spent % 60, (int)total_time / 3600, (int)total_time % 3600 / 60, (int)total_time % 60);
  }

  fprintf(stderr, "Beginning sampling...\n");

  for(k = params->seq_length - 1, i = 0; i < params->samples; i++) {
    params->acceptance_numer = params->acceptance_denom = 0;
    params->mutation_acceptance_numer = params->mutation_acceptance_denom = 0;
    params->branch_acceptance_numer = params->branch_acceptance_denom = 0;
    params->root_acceptance_numer = params->root_acceptance_denom = 0;
    params->rate_multiplier_acceptance_numer = params->rate_multiplier_acceptance_denom = 0;

    gettimeofday(&tv1, NULL);
    for(j = 0; j < params->sample_rate; j++) {
      if(k == params->seq_length - 1) {
	if(params->changes_out)
	  print_changes(params);
	if(params->variation) {
	  iterate_variation(params);
	  print_variation_parameters(params);
	}
	for(k = 0; k < params->seq_length; k++) {
	  r = random_int(k, params->seq_length - 1);
	  tmp = permutation[k];
	  permutation[k] = permutation[r];
	  permutation[r] = tmp;
	}
	k = 0;

	if(i) {
	  percent_complete += params->seq_length / total_iterations;
	  total_time = time_spent / percent_complete;
	  fprintf(stderr, "%2.3f%% complete, %.2d:%2.2d:%2.2ds elapsed (~%.2d:%2.2d:%2.2ds total)\n", percent_complete * 100.0, (int)time_spent / 3600, (int)time_spent % 3600 / 60, (int)time_spent % 60, (int)total_time / 3600, (int)total_time % 3600 / 60, (int)total_time % 60);
	}
      }
      else
	k++;
      if(params->root_position[permutation[j]] != -1) {
	iterate_gibbs(params, permutation[k], configuration, likelihood);
	if(random_double(0.0, 1.0) <= params->update_parameters_rate)
	  iterate_parameters(params);
	if(random_double(0.0, 1.0) <= params->update_root_rate && iterate_root_parameters(params))
	  update_root_table(params);
      }
    }
    normalize_rates(params);
    print_parameters(params);
    if(params->node_out) {
      count_nodes(params);
      if(i % 100 == 99)
	print_nodes(params);
    }

    gettimeofday(&tv2, NULL);
    tm = tv2.tv_sec - tv1.tv_sec + (tv2.tv_usec - tv1.tv_usec) / 1000000.0;
    time_spent += tm;
    if(params->stats)
      fprintf(stderr, "%d: change = %g, %g/sec; mutation = %g, %g/sec;\n   branch = %g %g/sec; root = %g, %g/sec; rate_multiplier = %g, %g/sec; seconds = %g\n", i, (double)params->acceptance_numer / params->acceptance_denom, params->acceptance_numer / tm, (double)params->mutation_acceptance_numer / params->mutation_acceptance_denom, params->mutation_acceptance_numer / tm, (double)params->branch_acceptance_numer / params->branch_acceptance_denom, params->branch_acceptance_numer / tm, (double)params->root_acceptance_numer / params->root_acceptance_denom, params->root_acceptance_numer / tm, (double)params->rate_multiplier_acceptance_numer / params->rate_multiplier_acceptance_denom, params->rate_multiplier_acceptance_numer / tm, tm);

  }
}

void iterate_parameters(Params* params) {
  if(random_double(0.0, 1.0) < 0.5) {
    iterate_mutation_parameters(params);
    if(params->variation)
      iterate_rate_multiplier_parameters(params);
    iterate_branch_parameters(params);
  }
  else {
    iterate_branch_parameters(params);
    if(params->variation)
      iterate_rate_multiplier_parameters(params);
    iterate_mutation_parameters(params);
  }
  update_branch_param_table(params);
}

void count_nodes(Params* params) {
  int i;

  for(i = 0; i < params->seq_length; i++)
    if(params->root_position[i] == 0)
      params->node_count[i].el[params->configurations[i][0]]++;
}

void print_nodes(Params* params) {
  int i, j;
  
  rewind(params->node_out);

  for(i = 0; i < params->seq_length; i++) {
    for(j = 0; j < NUM_NUC; j++)
      fprintf(params->node_out, " %u", (unsigned int)params->node_count[i].el[j]);
    fprintf(params->node_out, "\n");
  }
  fflush(params->node_out);
}

int get_tree_state(int tree_size, double* tree) {
  double r;
  int i;

  r = random_double(0.0, tree[1]);
  for(i = 2; i < tree_size; i *= 2)
    if(r > tree[i]) {
      r -= tree[i];
      i++;
    }
  return i / 2 - tree_size / 2;
}

void print_categories(Params* params) {
  int i, b;

  for(i = 0; i < params->seq_length; i++) {
    printf("cat[%d] = ", i);
    for(b = 0; b < params->num_branches; b++)
      printf(" %d", (int)params->branch_state[i][b].rate_category);
    printf("\n");
  }
  fflush(stdout);

}

void print_changes(Params* params) {
  int i, b, c, h;

  for(h = 0; h < params->num_hidden_states; h++)
    for(c = 0; c < params->num_rate_categories; c++)
      for(b = 0; b < params->num_branches; b++)
	if(params->category_branch_existence[c][b]) {
	  for(i = 0; i <  params->num_total_mutation_parameters[params->category_type[c]]; i++) 
	    fprintf(params->changes_out[c][b], " %5.1d", params->changes[h][c][b][i]);
	  fprintf(params->changes_out[c][b], "\n");
	  fflush(params->changes_out[c][b]);
	}
  
  /*
  for(c = 0; c < params->num_rate_categories; c++)
    for(b = 0; b < params->num_branches; b++) {
      sum = 0;
      printf("changes[%d] = ",  b);
      for(i = 0; i < params->num_total_mutation_parameters[params->category_type[c]]; i++) {
	printf("% 2.1d", params->changes[c][b][i]);
	sum += params->changes[c][b][i];
      }
      printf("\nsum = %d\n", sum);
    }

  for(c = 0; c < params->num_root_categories; c++) {
    printf("root_changes = ");
    for(i = 0; i < NUM_TOTAL_ROOT_PARAMETERS; i++)
      printf("% 4.1d", params->root_changes[c][i]);
    printf("\n");
  }
  

  fflush(stdout);
  */
}

void normalize_rates(Params* params) {
  int t, l, c, b, i, j, a;
  double freq;
  double scale, numer, denom;

  /*
  fprintf(stderr, ".");
  update_branch_param_table(params, 1);
  fprintf(stderr, ";");
  */
  for(l = 0; l < params->num_time_categories; l++) 
    for(t = 0; t < params->num_mutation_types; t++) {
      numer = denom = 0.0;
      for(c = 0; c < params->num_rate_categories; c++)
	if(params->rate_time_category[c] == l) {
	  a = params->category_type[c];
	  if(params->mutation_type_symmetry[t] == a || params->mutation_type_symmetry[t] == 2)
	    for(i = 0; i < params->num_mutation_parameters_per_type[a][t]; i++) {
	      j = params->mutation_parameters_per_type[a][t][i];
	      freq = params->trinuc_freqs[0][params->param_enum[a][j][0][0]][params->param_enum[a][j][0][1]][params->param_enum[a][j][0][2]];
	      numer += freq * params->parameters[c][j];
	      denom += freq;
	    }
	}
      scale = denom / numer;
      for(c = 0; c < params->num_rate_categories; c++)
	if(params->rate_time_category[c] == l) {
	  a = params->category_type[c];
	  if(params->mutation_type_symmetry[t] == a || params->mutation_type_symmetry[t] == 2)
	    for(i = 0; i < params->num_mutation_parameters_per_type[a][t]; i++) {
	      params->parameters[c][params->mutation_parameters_per_type[a][t][i]] *= scale;
	    }
	}
      scale = numer / denom;
      for(b = 0; b < params->num_branches; b++)
	if(params->time_category[b] == l) {
	  params->branch_ratio[b][t] *= scale;
	}
    }
  /*
  update_branch_param_table(params, 1);
  fprintf(stderr, ":");
  */
}


/*
void normalize_rates(Params* params) {
  int i, j, c, b;
  double* sum;
  int a;
  double* numer;
  int* denom;
  int* time;
  int tmp;

  (time = (int*)malloc(sizeof(int) * (params->num_total_mutation_parameters[1] - params->num_mutation_parameters[1]))) || die("Could not allocate memory for time\n");
  (numer = (double*)malloc(sizeof(double) * params->num_time_categories)) || die("Could not allocate memory for numer\n");
  (denom = (int*)malloc(sizeof(int) * params->num_time_categories)) || die("Could not allocate memory for denom\n");
  for(i = 0; i < params->num_time_categories; i++) {
    numer[i] = 0.0;
    denom[i] = 0;
  }

  for(c = 0; c < params->num_rate_categories; c++) {
    a = params->category_type[c];
    for(i = 0; i < params->num_total_mutation_parameters[a] - params->num_mutation_parameters[a]; i++) {
      time[i] = 0;
      for(b = 0; b < params->num_branches; b++)
	if(params->category_branch_existence[c][b])
	  for(j = 0; j < NUM_NUC; j++) {
	    if(j < 3)
	      tmp = params->param_enum_sum[a][i][j];
	    else
	      tmp = i + params->num_mutation_parameters[a];
	    time[i] += params->changes[c][b][tmp];
	  }
    }

    for(i = 0; i < params->num_mutation_parameters[a]; i++) {
      numer[params->rate_time_category[c]] += params->parameters[c][i] * time[params->param_enum_addend[a][i] - params->num_mutation_parameters[a]];
      denom[params->rate_time_category[c]] += time[params->param_enum_addend[a][i] - params->num_mutation_parameters[a]];
    }
  }
  free(time);

  for(b = 0; b < params->num_branches; b++)
    params->branch_ratio[b] *= numer[params->time_category[b]] / denom[params->time_category[b]];


  (sum = (double*)malloc(sizeof(double) * params->num_time_categories)) || die("Could not allocate memory for sum\n");
  for(i = 0; i < params->num_time_categories; i++)
    sum[i] = 0.0;
  for(b = 0; b < params->num_branches; b++)
    sum[params->time_category[b]] += params->branch_ratio[b] * params->branch_length[b];
  fprintf(stderr, "totaltime = ");
  for(i = 0; i < params->num_time_categories; i++)
    fprintf(stderr, " %g", sum[i]);
  fprintf(stderr, "\n");
  free(sum);

  for(c = 0; c < params->num_rate_categories; c++)
    for(i = 0; i < params->num_mutation_parameters[params->category_type[c]]; i++)
      params->parameters[c][i] *= denom[params->rate_time_category[c]] / numer[params->rate_time_category[c]];


  free(numer);
  free(denom);
}
*/

void print_parameters(Params* params) {
  int i, c, t;

  /*
  print_changes(params);
  count_configuration_changes(params);
  print_changes(params);
  */

  for(c = 0; c < params->num_rate_categories; c++) {
    for(i = 0; i < params->num_mutation_parameters[params->category_type[c]]; i++)
      fprintf(params->mutation_out[c], "%14.9g ", params->parameters[c][i]);
    fprintf(params->mutation_out[c], "\n");
    fflush(params->mutation_out[c]);
  }

  for(t = 0; t < params->num_mutation_types; t++) {
    for(i = 0; i < params->num_branches; i++)
      fprintf(params->branch_out[t], "%14.9g ", params->branch_ratio[i][t] * params->branch_length[i]);
    fprintf(params->branch_out[t], "\n");
    fflush(params->branch_out[t]);
  }   
 
  for(c = 0; c < params->num_root_categories; c++) {
    for(i = 0; i < NUM_ROOT_PARAMETERS; i++)
      fprintf(params->root_out[c], "%14.9g ", params->root_transitions[c][i]);
    fprintf(params->root_out[c], "\n");
    fflush(params->root_out[c]);
  }

  if(params->variation) {
    for(i = 0; i < params->num_hidden_states; i++) 
      fprintf(params->rate_multiplier_out, "%14.9g", params->rate_multiplier[i]);
    fprintf(params->rate_multiplier_out, "\n");
    fflush(params->rate_multiplier_out);
  }
}

void print_variation_parameters(Params* params) {
  int i, h;

  rewind(params->variation_out);
  for(i = 0; i < params->seq_length; i++) {
    for(h = 0; h < params->num_hidden_states; h++)
      fprintf(params->variation_out, " %u", (unsigned int)params->variation_count[i][h]);
    fprintf(params->variation_out, "\n");
  }
  fflush(params->variation_out);

  for(i = 0; i < params->num_hidden_states; i++)
    fprintf(params->initial_out, "%14.9g ", params->initial[i]);
  fprintf(params->initial_out, "\n");
  fflush(params->initial_out);

  for(i = 0; i < params->num_hidden_states; i++)
    for(h = 0; h < params->num_hidden_states; h++)
      fprintf(params->transition_out, "%15.9g ", params->transition[i][h]);
  fprintf(params->transition_out, "\n");
  fflush(params->transition_out);
}

/*
double calculate_probability(Params* params) {
  int h, c, b, i, a;
  double probability;

  probability = 0.0;
  
  for(h = 0; h < params->num_hidden_states; h++)
    for(c = 0; c < params->num_rate_categories; c++) {
      a = params->category_type[c];
      for(b = 0; b < params->num_branches; b++)      
	for(i = 0; i < params->num_total_mutation_parameters[a]; i++)
	  if(params->changes[h][c][b][i] > 0)
	    probability += log(params->branch_param_table[c][b].el[params->param_enum[a][i][0][0]][params->param_enum[a][i][0][1]][params->param_enum[a][i][0][2]][params->param_enum[a][i][0][3]]) * params->changes[c][b][i];
    }
  for(c = 0; c < params->num_root_categories; c++)
    for(i = 0; i < NUM_TOTAL_ROOT_PARAMETERS; i++)
      if(params->root_changes[c][i] > 0)
	probability += log(params->cond_root[c].el[params->root_enum[i][0]][params->root_enum[i][1]][params->root_enum[i][2]]) * params->root_changes[c][i];
  
  return probability;
}
*/

int iterate_mutation_parameters(Params* params) {
  double *parameters, *mean, *variance;
  double q_ratio, p_ratio, ratio;
  int h, i, j, b, c, tmpi, t;
  int a;
  double numer, denom, tmp2, tmp3, tmp4;
  int bad;


  /*
  print_changes(params);
  count_configuration_changes(params);
  print_changes(params);
  */

  (parameters = (double*)malloc(sizeof(double) * 256)) || die("Could not allocate memory for parameters\n");
  (mean = (double*)malloc(sizeof(double) * 256)) || die("Could not allocate memory for mean\n");
  (variance = (double*)malloc(sizeof(double) * 256)) || die("Could not allocate memory for variance\n");

  tmpi = params->mutation_acceptance_numer;

  for(c = 0; c < params->num_rate_categories; c++) {
    a = params->category_type[c];
    sample_mutation_parameters(params, c, parameters, mean, variance);
    for(i = 0; i < params->num_mutation_parameters[a]; i++) {
      if(parameters[i] > 0.0 && parameters[i] < params->max_rate) {
	tmp2 = 4.0 * variance[i];
	tmp3 = parameters[i] - mean[i];
	tmp4 = params->parameters[c][i] - mean[i];
	q_ratio = 2.5 * log((tmp2 + tmp3 * tmp3) / (tmp2 + tmp4 * tmp4));
	
	tmp2 = log(parameters[i] / params->parameters[c][i]);
	p_ratio = 0.0;
	
	for(h = 0; h < params->num_hidden_states; h++) {
	  for(b = 0; b < params->num_branches; b++) {
	    if(params->category_branch_existence[c][b]) {
	      numer = denom = 1.0;
	      for(j = 0; j < 3; j++) {
		tmpi = params->param_enum_sum[a][params->param_enum_addend[a][i] - params->num_mutation_parameters[a]][j];
		t = params->mutation_type[a][tmpi];
		denom -= params->branch_ratio[b][t] * params->parameters[c][tmpi] * params->rate_multiplier[h];
	      }
	      
	      t = params->mutation_type[a][i];
	      numer = denom - params->branch_ratio[b][t] * (-params->parameters[c][i] + parameters[i]) * params->rate_multiplier[h];
	      if(numer <= 0.0)
		break;
	      p_ratio += params->changes[h][c][b][i] * tmp2;
	      p_ratio += params->changes[h][c][b][params->param_enum_addend[a][i]] * log(numer / denom);
	    }
	  }
	  if(b < params->num_branches)
	    break;
	}
	if(h == params->num_hidden_states) {
	  /*
	  fprintf(stderr, "mutation: old[%d][%d] = %g, new = %g, mean = %g, variance = %g, ratio = %g + %g = %g\n", c, i, params->parameters[c][i], parameters[i], mean[i], variance[i], q_ratio, p_ratio, q_ratio + p_ratio);
	  */
	  ratio = q_ratio + p_ratio;
	  
	  if(ratio >= 0.0 || log(random_double(0.0, 1.0)) < ratio || !params->metropolis) {
	    params->parameters[c][i] = parameters[i];
	    params->mutation_acceptance_numer++;
	    }
	}
      }
      params->mutation_acceptance_denom++;
    }
  }
  
  free(parameters);
  free(mean);
  free(variance);

  return (tmpi != params->mutation_acceptance_numer);
}

void update_branch_param_table(Params* params) {
  int i, j, k, l,  b, c, a, p, h;
  FLOAT tmp, tmp2;

  for(c = 0; c < params->num_rate_categories; c++) {
    a = params->category_type[c];
    for(b = 0; b < params->num_branches; b++)
      if(params->category_branch_existence[c][b])
	for(i = 0; i < NUM_NUC; i++)
	  for(k = 0; k < NUM_NUC; k++) 
	    for(j = 0; j < NUM_NUC; j++) {
	      for(h = 0; h < params->num_hidden_states; h++)
	      params->branch_param_table[h][c][b].el[i][k][j][k] = 1.0;
	      
	      for(l = 0; l < NUM_NUC; l++)
		if(l != k) {
		  p = params->trinuc_parameters[a][64 * i + 16 * k + 4 * j + l];
		  tmp2 = params->branch_ratio[b][params->mutation_type[a][p]] * params->parameters[c][p];
		  for(h = 0; h < params->num_hidden_states; h++) {
		    tmp = params->branch_param_table[h][c][b].el[i][k][j][l] = tmp2 * params->rate_multiplier[h];
		    params->branch_param_table[h][c][b].el[i][k][j][k] -= tmp;
		    if(params->category_type[c])
		      params->reverse_branch_param_table[h][c][b].el[complement[j]][complement[k]][complement[i]][complement[l]] = tmp;
		  }
		}
	      if(params->category_type[c])
		for(h = 0; h < params->num_hidden_states; h++) 
		  params->reverse_branch_param_table[h][c][b].el[complement[j]][complement[k]][complement[i]][complement[k]] = params->branch_param_table[h][c][b].el[i][k][j][k];
	      /*
	      if(params->branch_param_table[c][b].el[i][k][j][k] <= 0.0)
		fprintf(stderr, "branch_param_table[%d][%d].el[%d][%d][%d][%d] = %g\n", c, b, i, k, j, k, params->branch_param_table[c][b].el[i][k][j][k]);
	      */
	    }
  }
}

void update_branch_param_list(Params* params) {
  int b, c, a, p, h;
  FLOAT tmp, tmp2;

  for(c = 0; c < params->num_rate_categories; c++) {
    a = params->category_type[c];
    for(b = 0; b < params->num_branches; b++)
      if(params->category_branch_existence[c][b]) {
	for(h = 0; h < params->num_hidden_states; h++)
	  for(p = params->num_mutation_parameters[a]; p < params->num_total_mutation_parameters[a]; p++)
	    params->branch_param_list[h][c][b][p] = 1.0;
	for(p = 0; p < params->num_mutation_parameters[a]; p++) {
	  tmp2 = params->branch_ratio[b][params->mutation_type[a][p]] * params->parameters[c][p];
	  for(h = 0; h < params->num_hidden_states; h++) {
	    tmp = params->branch_param_list[h][c][b][p] = tmp2 * params->rate_multiplier[h];
	    params->branch_param_list[h][c][b][params->param_enum_addend[a][p]] -= tmp;
	  }
	}
      }
  }
}

void sample_mutation_parameters(Params* params, int c, double* parameters, double* mle, double* variance) {
  int i;

  find_mutation_mle(params, c, mle, variance);

  for(i = 0; i < params->num_mutation_parameters[params->category_type[c]]; i++)
    /*    parameters[i] = random_normal(mle[i], variance[i] * params->alpha); */
    if(params->metropolis)
      parameters[i] = random_t(mle[i], variance[i], 4);
    else
      parameters[i] = mle[i];
}

void find_mutation_mle(Params* params, int c, double* parameters, double* variance) {
  int i, j, b, numer, tmp, t, h;
  double **denom, time;
  int a;

  a = params->category_type[c];
  (denom = (double**)malloc(sizeof(double) * params->num_mutation_types)) || die("Could not allocate memory for denom\n");
  for(t = 0; t < params->num_mutation_types; t++)
    (denom[t] = (double*)malloc(sizeof(double) * (params->num_total_mutation_parameters[a] - params->num_mutation_parameters[a]))) || die("Could not allocate memory for denom\n");
  
  for(t = 0; t < params->num_mutation_types; t++)
    for(i = 0; i < params->num_total_mutation_parameters[a] - params->num_mutation_parameters[a]; i++) {
      denom[t][i] = 0.0;
      for(h = 0; h < params->num_hidden_states; h++)
	for(b = 0; b < params->num_branches; b++) 
	  if(params->category_branch_existence[c][b]) {
	    time = 0.0;
	    for(j = 0; j < NUM_NUC; j++) {
	      if(j < 3)
		tmp = params->param_enum_sum[a][i][j];
	      else
		tmp = i + params->num_mutation_parameters[a];
	      time += params->changes[h][c][b][tmp];
	    }
	    denom[t][i] += time * params->branch_ratio[b][t] * params->rate_multiplier[h];
	  }
    }
  
  for(i = 0; i < params->num_mutation_parameters[a]; i++) {
    for(numer = h = 0; h < params->num_hidden_states; h++)
      for(b = 0; b < params->num_branches; b++)
	numer += params->changes[h][c][b][i];
    parameters[i] = numer / denom[params->mutation_type[a][i]][params->param_enum_addend[a][i] - params->num_mutation_parameters[a]];

    variance[i] = (parameters[i] * parameters[i]) / numer;
  }
  
  for(t = 0; t < params->num_mutation_types; t++)
    free(denom[t]);
  free(denom);
}
 
int iterate_branch_parameters(Params* params) {
  double* parameters[MAX_BRANCHES], *mean[MAX_BRANCHES], *variance[MAX_BRANCHES];
  double q_ratio, p_ratio, ratio, sum;
  int i, j, k, a, b, c, tmpi, h;
  double tmp2, tmp3, tmp4;
  double numer, denom;
  int type, it, t;
  int bad;

  for(b = 0; b < params->num_branches; b++) {
    (parameters[b] = (double*)malloc(sizeof(double) * params->num_mutation_types)) || die("Could not allocate memory for parameters\n");
    (mean[b] = (double*)malloc(sizeof(double) * params->num_mutation_types)) || die("Could not allocate memory for mean\n");
    (variance[b] = (double*)malloc(sizeof(double) * params->num_mutation_types)) || die("Could not allocate memory for variance\n");
  }

  tmpi = params->branch_acceptance_numer;

  sample_branch_parameters(params, parameters, mean, variance);
  
  for(b = 0; b < params->num_branches; b++) {
    for(t = 0; t < params->num_mutation_types; t++) {
      if(parameters[b][t] > 0.0 && parameters[b][t] < params->max_rate) {
	bad = 0;
	tmp2 = 4.0 * variance[b][t];
	tmp3 = parameters[b][t] - mean[b][t];
	tmp4 = params->branch_ratio[b][t] - mean[b][t];
	q_ratio = 2.5 * log((tmp2 + tmp3 * tmp3) / (tmp2 + tmp4 * tmp4));
	
	p_ratio = 0.0;
	for(h = 0; h < params->num_hidden_states; h++)
	  for(c = 0; c < params->num_rate_categories; c++)
	    if(params->category_branch_existence[c][b]) {
	      a = params->category_type[c];
	      if(params->mutation_type_symmetry[t] == a || params->mutation_type_symmetry[t] == 2)
		for(i = 0; i < params->num_mutation_parameters_per_type[a][t]; i++)
		  p_ratio += params->changes[h][c][b][params->mutation_parameters_per_type[a][t][i]];
	    }
	p_ratio *= log(parameters[b][t] / params->branch_ratio[b][t]);
	
	for(h = 0; !bad && h < params->num_hidden_states; h++)
	  for(c = 0; !bad && c < params->num_rate_categories; c++) 
	    if(params->category_branch_existence[c][b]) {
	      a = params->category_type[c];
	      if(params->mutation_type_symmetry[t] == a || params->mutation_type_symmetry[t] == 2) {
		for(i = 0; !bad && i < params->num_total_parameters_per_type[a][t]; i++) {
		  numer = denom = 0.0;
		  k = params->total_parameters_per_type[a][t][i];
		  for(j = 0; j < 3; j++) {
		    it = params->param_enum_sum[a][k][j];
		    type = params->mutation_type[a][it];
		    if(type == t)
		      numer += parameters[b][t] * params->parameters[c][it];
		    else
		      numer += params->branch_ratio[b][type] * params->parameters[c][it];
		    denom += params->branch_ratio[b][type] * params->parameters[c][it];
		  }
		  numer = 1.0 - numer * params->rate_multiplier[h];
		  denom = 1.0 - denom * params->rate_multiplier[h];
		  
		  if(numer <= 0.0) {
		    bad = 1;
		    break;
		  }
		  p_ratio += params->changes[h][c][b][k + params->num_mutation_parameters[a]] * log(numer / denom);
		}
	      }
	    }
	if(!bad) {
	  /*
	  fprintf(stderr, "branch: old[%d][%d] = %g, new = %g, mean = %g, variance = %g, ratio = %g + %g = %g\n", b, t, params->branch_ratio[b][t], parameters[b][t], mean[b][t], variance[b][t], q_ratio, p_ratio, q_ratio + p_ratio);
	  */
	  ratio = q_ratio + p_ratio;
	  if(ratio >= 0.0 || log(random_double(0.0, 1.0)) < ratio || !params->metropolis) {
	    params->branch_ratio[b][t] = parameters[b][t];
	    params->branch_acceptance_numer++;
	  }
	}
      }
      params->branch_acceptance_denom++;
    }
  }
  for(b = 0; b < params->num_branches; b++) {
    free(parameters[b]);
    free(mean[b]);
    free(variance[b]);
  }
  return (tmpi != params->branch_acceptance_numer);
}

void sample_branch_parameters(Params* params, double** parameters, double** mle, double** variance) {
  int i, t;
  
  find_branch_mle(params, mle, variance);

  for(i = 0; i < params->num_branches; i++)
    for(t = 0; t < params->num_mutation_types; t++)
      if(params->metropolis)
	parameters[i][t] = random_t(mle[i][t], variance[i][t], 4);
      else
	parameters[i][t] = mle[i][t];
}

void find_branch_mle(Params* params, double** parameters, double** variance) {
  int i, j, a, b, c, tmp, t, h;
  double numer, denom, ***time;
  
  (time = (double***)malloc(sizeof(double**) * params->num_hidden_states)) || die("Could not allocate memory for time\n");
  for(h = 0; h < params->num_hidden_states; h++) {
    (time[h] = (double**)malloc(sizeof(double*) * params->num_rate_categories)) || die("Could not allocate memory for time[%d]\n", h);
    for(c = 0; c < params->num_rate_categories; c++)
      (time[h][c] = (double*)malloc(sizeof(double) * (params->num_total_mutation_parameters[1] - params->num_mutation_parameters[1]))) || die("Could not allocate memory for time[%d][%d]\n", h, c);
  }
    
  for(b = 0; b < params->num_branches; b++) {
    for(h = 0; h < params->num_hidden_states; h++)
      for(c = 0; c < params->num_rate_categories; c++)
	if(params->category_branch_existence[c][b]) {
	  a = params->category_type[c];
	  for(i = 0; i < params->num_total_mutation_parameters[a] - params->num_mutation_parameters[a]; i++) {
	    time[h][c][i] = 0.0;
	    for(j = 0; j < NUM_NUC; j++) {
	      if(j < 3)
		tmp = params->param_enum_sum[a][i][j];
	      else
		tmp = i + params->num_mutation_parameters[a];
	      time[h][c][i] += params->changes[h][c][b][tmp];
	    }
	  }   
	}
    
    for(t = 0; t < params->num_mutation_types; t++) {
      numer = denom = 0.0;
      for(h = 0; h < params->num_hidden_states; h++)
	for(c = 0; c < params->num_rate_categories; c++)
	  if(params->category_branch_existence[c][b]) {
	    a = params->category_type[c];
	    if(params->mutation_type_symmetry[t] == a || params->mutation_type_symmetry[t] == 2)
	      for(i = 0; i < params->num_mutation_parameters_per_type[a][t]; i++) {
		tmp = params->mutation_parameters_per_type[a][t][i];
		numer += params->changes[h][c][b][tmp];
		denom += time[h][c][params->param_enum_addend[a][tmp] - params->num_mutation_parameters[a]] * params->parameters[c][i] * params->rate_multiplier[h];
	      }
	  }
      parameters[b][t] = numer / denom;
      /*
	fprintf(stderr, "parameters[%d] = %g = %g / %g\n", b, parameters[b], numer, denom);
      */
      variance[b][t] = (parameters[b][t] * parameters[b][t]) / numer;
    }
  }
  
  for(h = 0; h < params->num_hidden_states; h++) {
    for(c = 0; c < params->num_rate_categories; c++)
      free(time[h][c]);
    free(time[h]);
  }
  free(time);
}


int iterate_root_parameters(Params* params) {
  double parameters[NUM_ROOT_PARAMETERS], mean[NUM_ROOT_PARAMETERS];
  RootVariance variance;
  double q_ratio, p_ratio, ratio;
  int i, j, tmp, tmpi, tmpj;
  double lastp, last;
  double p[3], pp[3];
  double tmp2;
  int c;
  
  tmp = params->root_acceptance_numer;

  for(c = 0; c < params->num_root_categories; c++) {
    sample_root_parameters(params, c, parameters, mean, &variance);
    
#ifdef BLAH
    
    for(i = 0; i < NUM_ROOT_BASE; i++) {
      tmpi = params->root_enum[i + NUM_ROOT_PARAMETERS][0];
      tmpj = params->root_enum[i + NUM_ROOT_PARAMETERS][1];
      bad = 0;
      lastp = 1.0;
      for(j = 0; j < 3; j++) {
	if(parameters[params->root_parameters[tmpi][tmpj][j]] <= 0.0)
	  bad = 1;
	lastp -= parameters[params->root_parameters[tmpi][tmpj][j]];
      }
      if(!bad && lastp > 0.0) {
	for(j = 0; j < 3; j++) {
	  p[j] = params->root_transitions[params->root_parameters[tmpi][tmpj][j]] - mean[params->root_parameters[tmpi][tmpj][j]];
	  pp[j] = parameters[params->root_parameters[tmpi][tmpj][j]] - mean[params->root_parameters[tmpi][tmpj][j]];
	}
	/*
	  q_ratio = 0.0;
	  for(j = 0; j < 3; j++)
	  q_ratio += variance.inverse[i][j] * (pp[j] * pp[j] - p[j] * p[j]);
	  q_ratio *= 0.5;
	  q_ratio += variance.inverse[i][3] * (pp[0] * pp[1] + pp[0] * pp[2] + pp[1] * pp[2] - pp[0] * pp[1] - p[0] * p[2] - p[1] * p[2]);
	  q_ratio /= params->alpha;
	*/
	
	q_ratio = 0.0;
	for(j = 0; j < 3; j++)
	  q_ratio += 0.5 * (pp[j] * pp[j] - p[j] * p[j]) * variance.inverse[i][j] / params->alpha;
	
	last = 1.0;
	for(j = 0; j < 3; j++)
	  last -= params->root_transitions[params->root_parameters[tmpi][tmpj][j]];
	p_ratio = params->root_changes[tmpi][tmpj][3] * log(lastp / last);
	for(j = 0; j < 3; j++)
	p_ratio += params->root_changes[tmpi][tmpj][j] * log(parameters[params->root_parameters[tmpi][tmpj][j]] / params->root_transitions[params->root_parameters[tmpi][tmpj][j]]);
	ratio = q_ratio + p_ratio;
      /*
	fprintf(stderr, "i=%d, ratio = %g = %g + %g, new[%d]=%g,%g,%g old=%g,%g,%g mean=%g,%g,%g sd=%g,%g,%g\n", i, ratio, q_ratio, p_ratio, i, parameters[params->root_parameters[tmpi][tmpj][0]], parameters[params->root_parameters[tmpi][tmpj][1]], parameters[params->root_parameters[tmpi][tmpj][2]], params->root_transitions[params->root_parameters[tmpi][tmpj][0]], params->root_transitions[params->root_parameters[tmpi][tmpj][1]], params->root_transitions[params->root_parameters[tmpi][tmpj][2]], mean[params->root_parameters[tmpi][tmpj][0]], mean[params->root_parameters[tmpi][tmpj][1]], mean[params->root_parameters[tmpi][tmpj][2]], sqrt(variance.el[i][0][0]),  sqrt(variance.el[i][1][1]), sqrt(variance.el[i][2][2]));
      */
	if(ratio >= 0.0 || log(random_double(0.0, 1.0)) < ratio || !params->metropolis) {
	  for(j = 0; j < 3; j++)
	    params->root_transitions[params->root_parameters[tmpi][tmpj][j]] = parameters[params->root_parameters[tmpi][tmpj][j]];
	params->root_acceptance_numer++;
	}
      }
    /*
      else
      fprintf(stderr, "new[%d]=%g,%g,%g old=%g,%g,%g mean=%g,%g,%g\n", i, parameters[params->root_parameters[tmpi][tmpj][0]], parameters[params->root_parameters[tmpi][tmpj][1]], parameters[params->root_parameters[tmpi][tmpj][2]], params->root_transitions[params->root_parameters[tmpi][tmpj][0]], params->root_transitions[params->root_parameters[tmpi][tmpj][1]], params->root_transitions[params->root_parameters[tmpi][tmpj][2]], mean[params->root_parameters[tmpi][tmpj][0]], mean[params->root_parameters[tmpi][tmpj][1]], mean[params->root_parameters[tmpi][tmpj][2]]);
    */
      
      params->root_acceptance_denom++;
    }
#endif

    
    for(i = 0; i < NUM_ROOT_BASE; i++) {
      tmpi = params->root_enum[i + NUM_ROOT_PARAMETERS][0];
      tmpj = params->root_enum[i + NUM_ROOT_PARAMETERS][1];
      
      last = 1.0;
      for(j = 0; j < 3; j++)
	last -= params->root_transitions[c][params->root_parameters[tmpi][tmpj][j]];
      
      for(j = 0; j < 3; j++) {
	lastp = last + params->root_transitions[c][params->root_parameters[tmpi][tmpj][j]] - parameters[params->root_parameters[tmpi][tmpj][j]];
	if(parameters[params->root_parameters[tmpi][tmpj][j]] > 0.0  && lastp > 0.0) {
	  p[j] = params->root_transitions[c][params->root_parameters[tmpi][tmpj][j]] - mean[params->root_parameters[tmpi][tmpj][j]];
	  pp[j] = parameters[params->root_parameters[tmpi][tmpj][j]] - mean[params->root_parameters[tmpi][tmpj][j]];
	  
	  /*	q_ratio = 0.5 * (pp[j] * pp[j] - p[j] * p[j]) * variance.inverse[i][j] / params->alpha; */
	  tmp2 = 4.0 / variance.inverse[i][j];
	  q_ratio = 2.5 * log((tmp2 + pp[j] * pp[j]) / (tmp2 + p[j] * p[j]));
	
	  p_ratio = params->root_changes[c][params->root_parameters[tmpi][tmpj][3]] * log(lastp / last);
	  p_ratio += params->root_changes[c][params->root_parameters[tmpi][tmpj][j]] * log(parameters[params->root_parameters[tmpi][tmpj][j]] / params->root_transitions[c][params->root_parameters[tmpi][tmpj][j]]);
	  ratio = q_ratio + p_ratio;
	  /*
	  fprintf(stderr, "i=%d, ratio = %g = %g + %g, new[%d]=%g,%g,%g old=%g,%g,%g mean=%g,%g,%g sd=%g,%g,%g\n", i, ratio, q_ratio, p_ratio, i, parameters[params->root_parameters[tmpi][tmpj][0]], parameters[params->root_parameters[tmpi][tmpj][1]], parameters[params->root_parameters[tmpi][tmpj][2]], (double)params->root_transitions[c][params->root_parameters[tmpi][tmpj][0]], (double)params->root_transitions[c][params->root_parameters[tmpi][tmpj][1]], (double)params->root_transitions[c][params->root_parameters[tmpi][tmpj][2]], mean[params->root_parameters[tmpi][tmpj][0]], mean[params->root_parameters[tmpi][tmpj][1]], mean[params->root_parameters[tmpi][tmpj][2]], sqrt(variance.el[i][0][0]),  sqrt(variance.el[i][1][1]), sqrt(variance.el[i][2][2]));
	  */
	  if(ratio >= 0.0 || log(random_double(0.0, 1.0)) < ratio) {
	  params->root_transitions[c][params->root_parameters[tmpi][tmpj][j]] = parameters[params->root_parameters[tmpi][tmpj][j]];
	  last = lastp;
	  params->root_acceptance_numer++;
	  }
	}
	/*
	  else
	  fprintf(stderr, "new[%d]=%g,%g,%g old=%g,%g,%g mean=%g,%g,%g\n", i, parameters[params->root_parameters[tmpi][tmpj][0]], parameters[params->root_parameters[tmpi][tmpj][1]], parameters[params->root_parameters[tmpi][tmpj][2]], params->root_transitions[params->root_parameters[tmpi][tmpj][0]], params->root_transitions[params->root_parameters[tmpi][tmpj][1]], params->root_transitions[params->root_parameters[tmpi][tmpj][2]], mean[params->root_parameters[tmpi][tmpj][0]], mean[params->root_parameters[tmpi][tmpj][1]], mean[params->root_parameters[tmpi][tmpj][2]]);
	*/
	
	params->root_acceptance_denom++;
      }
    }
  }

  return(tmp != params->root_acceptance_numer);
}

void update_root_table(Params* params) {
  int i, j, k, c;
  double prob[2][NUM_NUC][NUM_NUC];
  double sum;
  int t, m, s;

  for(c = 0; c < params->num_root_categories; c++) {
    for(i = 0; i < NUM_NUC; i++)
      for(j = 0; j < NUM_NUC; j++) {
	params->cond_root[c].el[i][j][NUM_NUC - 1] = 1.0;
	for(k = 0; k < NUM_NUC - 1; k++) {
	  params->cond_root[c].el[i][j][k] = params->root_transitions[c][params->root_parameters[i][j][k]];
	  params->cond_root[c].el[i][j][NUM_NUC - 1] -= params->cond_root[c].el[i][j][k];
	}
      }

    for(i = 0; i < NUM_NUC; i++)
      for(j = 0; j < NUM_NUC; j++)
	prob[0][i][j] = 1.0 / 16.0;
    for(m = 0; m < 100; m++) {
      s = m % 2;
      t = (m+1) % 2;
      sum = 0.0;
      for(j = 0; j < NUM_NUC; j++)
	for(k = 0; k < NUM_NUC; k++) {
	  prob[t][j][k] = 0;
	  for(i = 0; i < NUM_NUC; i++)
	    prob[t][j][k] += prob[s][i][j] * params->cond_root[c].el[i][j][k];
	  sum += prob[t][j][k];
	}
      for(j = 0; j < NUM_NUC; j++)
	for(k = 0; k < NUM_NUC; k++)
	  prob[t][j][k] /= sum;
    }
    
    for(i = 0; i < NUM_NUC; i++)
      for(j = 0; j < NUM_NUC; j++)
	for(k = 0; k < NUM_NUC; k++)
	  params->reverse_cond_root[c].el[i][j][k] = params->cond_root[c].el[complement[k]][complement[j]][complement[i]] * prob[1][complement[k]][complement[j]] / prob[1][complement[j]][complement[i]];

  }
}

void sample_root_parameters(Params* params, int c, double* parameters, double* mle, RootVariance* variance) {
  double cholesky[3][3], sum, r[3];
  int i, j, k, l, tmpi, tmpj;
  
  find_root_mle(params, c,  mle, variance);

  /*
  for(i = 0; i < NUM_ROOT_BASE; i++) {
    for(j = 0; j < 3; j++)
      for(k = j; k < 3; k++) {
	sum = variance->el[i][j][k];
	for(l = 0; l < j; l++)
	  sum -= cholesky[j][l] * cholesky[k][l]; 
	if(j == k) {
	  if(sum <= 0.0)
	    die("Error: choleksy failed\n");
	  cholesky[j][j] = sqrt(sum);
	}
	else
	  cholesky[k][j] = sum / cholesky[j][j];
      }
    for(j = 0; j < 3; j++)
      r[j] = random_normal(0.0, params->alpha);
    tmpi = params->root_enum[i + NUM_ROOT_PARAMETERS][0];
    tmpj = params->root_enum[i + NUM_ROOT_PARAMETERS][1];
    for(j = 0; j < 3; j++) {
      parameters[params->root_parameters[tmpi][tmpj][j]] = mle[params->root_parameters[tmpi][tmpj][j]];
      for(k = 0; k <= j; k++)
	parameters[params->root_parameters[tmpi][tmpj][j]] += cholesky[j][k] * r[k];
    }
  */
  /*
    for(j = 0; j < 3; j++) {
      for(k = 0; k < 3; k++)
	fprintf(stderr, "%g,%g   ", variance->el[i][j][k], cholesky[j][k]);
      fprintf(stderr, "\n");
    }
    fprintf(stderr, "\n");

  }
  */

  for(i = 0; i < NUM_ROOT_BASE; i++) {
    tmpi = params->root_enum[i + NUM_ROOT_PARAMETERS][0];
    tmpj = params->root_enum[i + NUM_ROOT_PARAMETERS][1];
    for(j = 0; j < 3; j++)
      /*      parameters[params->root_parameters[tmpi][tmpj][j]] = random_normal(mle[params->root_parameters[tmpi][tmpj][j]], params->alpha / variance->inverse[i][j]); */
      if(params->metropolis)
	parameters[params->root_parameters[tmpi][tmpj][j]] = random_t(mle[params->root_parameters[tmpi][tmpj][j]], 1.0 / variance->inverse[i][j], 4);
      else
	parameters[params->root_parameters[tmpi][tmpj][j]] = mle[params->root_parameters[tmpi][tmpj][j]];
  }
}

void find_root_mle(Params* params, int c, double* parameters, RootVariance* variance) {
  int sum[NUM_ROOT_BASE], tmp, tmpi, tmpj;
  double sum_parameters[NUM_ROOT_BASE];
  double a[NUM_NUC], denom;
  int i, j;
  
  for(i = 0; i < NUM_ROOT_BASE; i++) {
    sum[i] = 0;
    for(j = 0; j < NUM_NUC; j++) {
      if(j < 3)
	tmp = params->root_enum_sum[i][j];
      else
	tmp = i + NUM_ROOT_PARAMETERS;
      sum[i] += params->root_changes[c][params->root_parameters[params->root_enum[tmp][0]][params->root_enum[tmp][1]][params->root_enum[tmp][2]]];
    }
  }

  for(i = 0; i < NUM_ROOT_BASE; i++)
    sum_parameters[i] = 1.0;
  for(i = 0; i < NUM_ROOT_PARAMETERS; i++) {
    parameters[i] = (double)params->root_changes[c][params->root_parameters[params->root_enum[i][0]][params->root_enum[i][1]][params->root_enum[i][2]]] / sum[params->root_enum_addend[i] - NUM_ROOT_PARAMETERS];
    sum_parameters[params->root_enum_addend[i] - NUM_ROOT_PARAMETERS] -= parameters[i];
  }

  for(i = 0; i < NUM_ROOT_BASE; i++) {
    tmpi = params->root_enum[i + NUM_ROOT_PARAMETERS][0];
    tmpj = params->root_enum[i + NUM_ROOT_PARAMETERS][1];
    a[3] = params->root_changes[c][params->root_parameters[tmpi][tmpj][3]] / (sum_parameters[i] * sum_parameters[i]);
    for(j = 0; j < 3; j++)
      a[j] = a[3] + params->root_changes[c][params->root_parameters[tmpi][tmpj][j]] / (parameters[params->root_parameters[tmpi][tmpj][j]] * parameters[params->root_parameters[tmpi][tmpj][j]]);
    denom = a[0] * a[1] * a[2] - (a[0] + a[1] + a[2]) * a[3] * a[3] + 2.0 * a[3] * a[3] * a[3];
    variance->el[i][0][0] = (a[1] * a[2] - a[3] * a[3]) / denom;
    variance->el[i][1][1] = (a[0] * a[2] - a[3] * a[3]) / denom;
    variance->el[i][2][2] = (a[0] * a[1] - a[3] * a[3]) / denom;
    variance->el[i][1][0] = variance->el[i][0][1] = a[3] * (a[3] - a[2]) / denom;
    variance->el[i][2][0] = variance->el[i][0][2] = a[3] * (a[3] - a[1]) / denom;
    variance->el[i][1][2] = variance->el[i][2][1] = a[3] * (a[3] - a[0]) / denom;
    for(j = 0; j < NUM_NUC; j++)
      variance->inverse[i][j] = a[j];
  }
}

int iterate_rate_multiplier_parameters(Params* params) {
  double *parameters, *mean, *variance;
  double q_ratio, p_ratio, ratio;
  int h, i, j, b, c, tmpi, t;
  int a;
  double numer, denom, tmp2, tmp3, tmp4;
  int bad;

  (parameters = (double*)malloc(sizeof(double) * params->num_hidden_states)) || die("Could not allocate memory for parameters\n");
  (mean = (double*)malloc(sizeof(double) * params->num_hidden_states)) || die("Could not allocate memory for mean\n");
  (variance = (double*)malloc(sizeof(double) * params->num_hidden_states)) || die("Could not allocate memory for variance\n");

  tmpi = params->rate_multiplier_acceptance_numer;

  sample_rate_multiplier_parameters(params, parameters, mean, variance);
  for(h = 0; h < params->num_hidden_states; h++) 
    if(!params->fix_rate_multiplier[h]) {
      if(parameters[h] > 0.0 && parameters[h] < params->max_rate) {
	bad = 0;
	tmp2 = 4.0 * variance[h];
	tmp3 = parameters[h] - mean[h];
	tmp4 = params->rate_multiplier[h] - mean[h];
	q_ratio = 2.5 * log((tmp2 + tmp3 * tmp3) / (tmp2 + tmp4 * tmp4));
	
	tmp2 = log(parameters[h] / params->rate_multiplier[h]);
	p_ratio = 0.0;
	
	for(c = 0; c < !bad && params->num_rate_categories; c++) {
	  a = params->category_type[c];
	  for(b = 0; !bad && b < params->num_branches; b++) 
	    if(params->category_branch_existence[c][b]) {
	      for(i = 0; i < params->num_total_mutation_parameters[a] - params->num_mutation_parameters[a]; i++) {
		numer = denom = 1.0;
		for(j = 0; j < 3; j++) {
		  tmpi = params->param_enum_sum[a][i][j];
		  t = params->mutation_type[a][tmpi];
		  denom -= params->branch_ratio[b][t] * params->parameters[c][tmpi] * params->rate_multiplier[h];
		  numer -= params->branch_ratio[b][t] * params->parameters[c][tmpi] * parameters[h];
		}
		if(numer <= 0.0) {
		  bad = 1;
		  break;
		}
		p_ratio += params->changes[h][c][b][params->num_mutation_parameters[a] + i] * log(numer / denom);
	      }
	      for(i = 0; i < params->num_mutation_parameters[a]; i++)
		p_ratio += params->changes[h][c][b][i] * tmp2;
	    }
	}
	if(!bad) {
	  /*
	  fprintf(stderr, "rate_multiplier: old[%d] = %g, new = %g, mean = %g, variance = %g, ratio = %g + %g = %g\n", h, params->rate_multiplier[h], parameters[h], mean[h], variance[h], q_ratio, p_ratio, q_ratio + p_ratio);
	  */
	  ratio = q_ratio + p_ratio;
	  
	  if(ratio >= 0.0 || log(random_double(0.0, 1.0)) < ratio || !params->metropolis) {
	    params->rate_multiplier[h] = parameters[h];
	    params->rate_multiplier_acceptance_numer++;
	  }
	}
      }
      params->rate_multiplier_acceptance_denom++;
    }
  
  free(parameters);
  free(mean);
  free(variance);
  
  return (tmpi != params->rate_multiplier_acceptance_numer);
}


void sample_rate_multiplier_parameters(Params* params, double* parameters, double* mle, double* variance) {
  int i;

  find_rate_multiplier_mle(params, mle, variance);

  for(i = 0; i < params->num_hidden_states; i++)
    if(!params->fix_rate_multiplier[i]) {
      if(params->metropolis)
	parameters[i] = random_t(mle[i], variance[i], 4);
      else
	parameters[i] = mle[i];
    }
}


void find_rate_multiplier_mle(Params* params, double* parameters, double* variance) {
  int i, j, b, numer, tmp, h, c;
  double denom, *time;
  int a;

  (time = (double*)malloc(sizeof(double) * (params->num_total_mutation_parameters[1] - params->num_mutation_parameters[1]))) || die("Could not allocate memory for time\n");

  for(h = 0; h < params->num_hidden_states; h++) 
    if(!params->fix_rate_multiplier[h]) {
      numer = 0;
      denom = 0.0;
      for(c = 0; c < params->num_rate_categories; c++) {
	a = params->category_type[c];
	for(b = 0; b < params->num_branches; b++) 
	  if(params->category_branch_existence[c][b]) {
	    
	    for(i = 0; i < params->num_total_mutation_parameters[a] - params->num_mutation_parameters[a]; i++) {
	      time[i] = 0.0;
	      for(j = 0; j < NUM_NUC; j++) {
		if(j < 3)
		  tmp = params->param_enum_sum[a][i][j];
		else
		  tmp = i + params->num_mutation_parameters[a];
		time[i] += params->changes[h][c][b][tmp];
	      }
	    }
	    
	    for(i = 0; i < params->num_mutation_parameters[a]; i++) {
	      numer += params->changes[h][c][b][i];
	      denom += time[params->param_enum_addend[a][i] - params->num_mutation_parameters[a]] * params->parameters[c][i] * params->branch_ratio[b][params->mutation_type[a][i]];
	      /*
	      fprintf(stderr, "time[%d] = %g, parameters[%d][%d] = %g, branch_ratio[%d][%d] = %g\n", params->param_enum_addend[a][i] - params->num_mutation_parameters[a], time[params->param_enum_addend[a][i] - params->num_mutation_parameters[a]], c, i, params->parameters[c][i], b, params->mutation_type[a][i], params->branch_ratio[b][params->mutation_type[a][i]]);
	      */
	    }
	  }
      }
      /*
      fprintf(stderr, "parameters[%d] = %d / %g = %g\n", h, numer, denom, numer / denom);
      */
      parameters[h] = numer / denom;
      variance[h] = (parameters[h] * parameters[h]) / numer;
    }

  free(time);
}
  
void iterate_gibbs(Params* params, int pos, char* configuration, FLOAT** likelihood) {
  ConfigP configp;
  int changes;

  memcpy(configuration, params->configurations[pos], sizeof(char) * params->tree_time);

  configp.configuration = params->configurations[pos];
  configp.old_config = configuration;
  configp.pos = pos;

  conditional_likelihood(params, &configp, likelihood);
  /*
  fprintf(stderr, "%g %g %g %g\n", (double)likelihood[0][0], (double)likelihood[0][1], (double)likelihood[0][2], (double)likelihood[0][3]);
  */
  changes = gibbs_sample(params, &configp, likelihood);
  if(changes) { 
    if(params->record)
      count_changes(params, &configp);
    params->acceptance_numer++;
  }

  params->acceptance_denom++;
}

void count_changes(Params* params, ConfigP* configp) {
  int i, im1;
  int* changes, *changes_left, *changes_right, *node_changes;
  BranchState* branch_state, *branch_state_im1, *branch_state_left, *branch_state_right, *branch_state_left_im1, *branch_state_right_im1;
  /*  char cm1, cm2, cp1, cp2, c, old; */
  char* pcm1, *pcm2, *pcp1, *pcp2;
  char* pcm1_im1, *pcm2_im1, *pcp1_im1, *pcp2_im1;
  char* pc, *pc_old, *pc_im1, *pc_old_im1;
  int* root_parameters, *trinuc_parameters, *trinuc_parameters_left, *trinuc_parameters_right;
  Node* node;
  int left, right, right2;
  int axil2;
  int* root_count;
  int hidden_state;

  hidden_state = params->hidden_state[configp->pos];

  (root_count = (int*)malloc(sizeof(int) * (params->num_branches + 1))) || die("Could not allocate memory for root_count\n");
  for(i = 0; i <= params->num_branches; i++)
    root_count[i] = 0;
  /*
  if(configp->configuration[0] != configp->old_config[0])
    fprintf(stderr, "need root change at %d, %c%c(%c%c)%c%c->%c(%c%c)%c\n", configp->pos, NAUC[params->configurations[configp->pos - 2][0]], NAUC[params->configurations[configp->pos - 1][0]], NAUC[configp->old_config[0]], NAUC[configp->configuration[0]], NAUC[params->configurations[configp->pos + 1][0]], NAUC[params->configurations[configp->pos + 2][0]], NAUC[params->configurations[configp->pos - 1][1]], NAUC[configp->old_config[1]], NAUC[configp->configuration[1]], NAUC[params->configurations[configp->pos + 1][1]]);
  */
  for(i = params->root_position[configp->pos] + 1; i < params->tree_time; ++i) {
    if(params->configuration_map[i].type == AXIL) {
      axil2 = 1;
      changes = NULL;
      node = params->configuration_map[i].node;
      branch_state = params->branch_state[configp->pos] + node->num;

      im1 = node->parent->configuration_index + node->parent->time_to_parent - 1;
      branch_state_im1 = params->branch_state[configp->pos] + node->parent->num;

      if(!branch_state_im1->present) {
	i += node->time_to_parent - 1;
	continue;
      }

      pc_im1 = configp->configuration + im1;
      pc_old_im1 = configp->old_config + im1;

      left = branch_state_im1->neighbors ? branch_state_im1->neighbors->node_left : (configp->pos - 1);
      right = branch_state_im1->neighbors ? branch_state_im1->neighbors->node_right : (configp->pos + 1);
	
      if(*pc_im1 != *pc_old_im1) {
	pcm1_im1 = params->configurations[left] + im1;
	pcp1_im1 = params->configurations[right] + im1;

	branch_state_left = params->branch_state[left] + node->num;
	branch_state_right = params->branch_state[right] + node->num;

	if(branch_state_left->present == 1) {
	  changes_left = params->changes[params->hidden_state[left]][branch_state_left->rate_category][node->num];
	  trinuc_parameters_left = (branch_state_left->direction >= 0 ? params->trinuc_parameters[branch_state_left->direction] : params->complement_trinuc_parameters);
	  pcm1 = params->configurations[left] + i;
	  branch_state_left_im1 = params->branch_state[left] + node->parent->num;
	  pcm2_im1 = params->configurations[branch_state_left_im1->neighbors ? branch_state_left_im1->neighbors->node_left : (left - 1)] + im1;
	  /*
	  if(im1 == 0)
	    fprintf(stderr, "%c%c(%c%c)->%c\n", NAUC[*pcm2_im1], NAUC[*pcm1_im1], NAUC[*pc_old_im1], NAUC[*pc_im1], NAUC[*pcm1]);
	  */
	  --changes_left[trinuc_parameters_left[64 * *pcm2_im1 + 16 * *pcm1_im1 + 4 * *pc_old_im1 + *pcm1]];
	  ++changes_left[trinuc_parameters_left[64 * *pcm2_im1 + 16 * *pcm1_im1 + 4 * *pc_im1 + *pcm1]];
	}
	if(branch_state_right->present == 1) {
	  changes_right = params->changes[params->hidden_state[right]][branch_state_right->rate_category][node->num];
	  trinuc_parameters_right = (branch_state_right->direction >= 0 ? params->trinuc_parameters[branch_state_right->direction] : params->complement_trinuc_parameters);
	  pcp1 = params->configurations[right] + i;
	  branch_state_right_im1 = params->branch_state[right] + node->parent->num;
	  pcp2_im1 = params->configurations[branch_state_right_im1->neighbors ? branch_state_right_im1->neighbors->node_right : (right + 1)] + im1;
	  
	  --changes_right[trinuc_parameters_right[64 * *pc_old_im1 + 16 * *pcp1_im1 + 4 * *pcp2_im1 + *pcp1]];
	  ++changes_right[trinuc_parameters_right[64 * *pc_im1 + 16 * *pcp1_im1 + 4 * *pcp2_im1 + *pcp1]];
	}

	if(!root_count[node->parent->num]) {
	  root_count[node->parent->num] = 1;
	  if(params->root_position[right] == im1) {
	    node_changes = params->root_changes[params->root_category[right]];
	    branch_state_right = params->branch_state[right] + node->parent->num;
	    root_parameters = (branch_state_right->direction >= 0 ? params->root_parameters_vector : params->complement_root_parameters_vector);
	    
	    --node_changes[root_parameters[16 * *pcm1_im1 + 4 * *pc_old_im1 + *pcp1_im1]];
	    ++node_changes[root_parameters[16 * *pcm1_im1 + 4 * *pc_im1 + *pcp1_im1]];
	  }
	  if(params->root_position[configp->pos] == im1) {
	    node_changes = params->root_changes[params->root_category[configp->pos]];
	    root_parameters = (branch_state->direction >= 0 ? params->root_parameters_vector : params->complement_root_parameters_vector);
	    
	    branch_state_left_im1 = params->branch_state[left] + node->parent->num;
	    pcm2_im1 = params->configurations[branch_state_left_im1->neighbors ? branch_state_left_im1->neighbors->node_left : (left - 1)] + im1;

	    --node_changes[root_parameters[16 * *pcm2_im1 + 4 * *pcm1_im1 + *pc_old_im1]];
	    ++node_changes[root_parameters[16 * *pcm2_im1 + 4 * *pcm1_im1 + *pc_im1]];
	  }
	  branch_state_right_im1 = params->branch_state[right] + node->parent->num;
	  right2 = branch_state_right_im1->neighbors ? branch_state_right_im1->neighbors->node_right : (right + 1);
	  if(params->root_position[right2] == im1) {
	    node_changes = params->root_changes[params->root_category[right2]];
	    branch_state_right = params->branch_state[right2] + node->parent->num;
	    root_parameters = (branch_state_right->direction >= 0 ? params->root_parameters_vector : params->complement_root_parameters_vector);
	    pcp2_im1 = params->configurations[right2] + im1;
	    
	    --node_changes[root_parameters[16 * *pc_old_im1 + 4 * *pcp1_im1 + *pcp2_im1]];
	    ++node_changes[root_parameters[16 * *pc_im1 + 4 * *pcp1_im1 + *pcp2_im1]];
	  }
	}
      }
      
      if(branch_state->present != 1) {
	i += node->time_to_parent - 1;
	continue;
      }
      
      changes = params->changes[hidden_state][branch_state->rate_category][node->num];
      trinuc_parameters = (branch_state->direction >= 0 ? params->trinuc_parameters[branch_state->direction] : params->complement_trinuc_parameters);
	
      pc = configp->configuration + i;
      pc_old = configp->old_config + i;
      
      if(*pc != *pc_old || *pc_im1 != *pc_old_im1) {
	if(*pc_im1 == *pc_old_im1) {
	  pcm1_im1 = params->configurations[left] + im1;
	  pcp1_im1 = params->configurations[right] + im1;
	}
	--changes[trinuc_parameters[64 * *pcm1_im1 + 16 * *pc_old_im1 + 4 * *pcp1_im1 + *pc_old]];
	++changes[trinuc_parameters[64 * *pcm1_im1 + 16 * *pc_im1 + 4 * *pcp1_im1 + *pc]];
      }
    }
    else {
      changes || die("changes not properly defined: pos=%d, node->num=%d, i=%d\n", configp->pos, node->num, i);
      if(axil2) {
	left = branch_state->neighbors ? branch_state->neighbors->left : (configp->pos - 1);
	right = branch_state->neighbors ? branch_state->neighbors->right : (configp->pos + 1);
	branch_state_left = params->branch_state[left] + node->num;
	branch_state_right = params->branch_state[right] + node->num;
	changes_left = params->changes[params->hidden_state[left]][params->branch_state[left][node->num].rate_category][node->num];
	changes_right = params->changes[params->hidden_state[right]][params->branch_state[right][node->num].rate_category][node->num];
	trinuc_parameters_left = (branch_state_left->direction >= 0 ? params->trinuc_parameters[branch_state_left->direction] : params->complement_trinuc_parameters);
	trinuc_parameters_right = (branch_state_right->direction >= 0 ? params->trinuc_parameters[branch_state_right->direction] : params->complement_trinuc_parameters);
	pcm1 = params->configurations[left] + i;
	pcp1 = params->configurations[right] + i;
	pcm2 = params->configurations[branch_state_left->neighbors ? branch_state_left->neighbors->left : (left - 1)] + i;
	pcp2 = params->configurations[branch_state_right->neighbors ? branch_state_right->neighbors->right : (right + 1)] + i;
	pcm1_im1 = pcm1 - 1;
	pcp1_im1 = pcp1 - 1;
	pcm2_im1 = pcm2 - 1;
	pcp2_im1 = pcp2 - 1;
	
	pc_im1 = pc;
	pc_old_im1 = pc_old;
	++pc;
	++pc_old;

	axil2 = 0;
      }
      else {
	pcm1_im1 = pcm1;
	pcp1_im1 = pcp1;
	pcm2_im1 = pcm2;
	pcp2_im1 = pcp2;
	
	pc_im1 = pc;
	pc_old_im1 = pc_old;
	
	++pcm1;
	++pcp1;
	++pcm2;
	++pcp2;
	++pc;
	++pc_old;
      }
      if(*pc != *pc_old || *pc_im1 != *pc_old_im1) {
	--changes[trinuc_parameters[64 * *pcm1_im1 + 16 * *pc_old_im1 + 4 * *pcp1_im1 + *pc_old]];
	++changes[trinuc_parameters[64 * *pcm1_im1 + 16 * *pc_im1 + 4 * *pcp1_im1 + *pc]];
	if(*pc_im1 != *pc_old_im1) {
	  --changes_left[trinuc_parameters_left[64 * *pcm2_im1 + 16 * *pcm1_im1 + 4 * *pc_old_im1 + *pcm1]];
	  ++changes_left[trinuc_parameters_left[64 * *pcm2_im1 + 16 * *pcm1_im1 + 4 * *pc_im1 + *pcm1]];
	  --changes_right[trinuc_parameters_right[64 * *pc_old_im1 + 16 * *pcp1_im1 + 4 * *pcp2_im1 + *pcp1]];
	  ++changes_right[trinuc_parameters_right[64 * *pc_im1 + 16 * *pcp1_im1 + 4 * *pcp2_im1 + *pcp1]];
	}
      }
    }
  }
  free(root_count);
}

int gibbs_sample(Params* params, ConfigP* configp, FLOAT** likelihood) {
  int i, j, im1;
  double vec[NUM_NUC];
  int changes = 0;
  FLOAT *tmpp, *tmpp2;
  double *tmpp3, *tmpp4;
  char* pcm1, *pcp1;
  char* pcm1_im1, *pcp1_im1;
  char* pc, *pc_old, *pc_im1, *pc_old_im1;
  BranchState* branch_state, *branch_state_im1;
  ParamTable* param_table = NULL;
  int hidden_state;

  hidden_state = params->hidden_state[configp->pos];
  i = params->root_position[configp->pos];
  pc = configp->configuration + i;
  pc_old = configp->old_config + i;
  tmpp2 = likelihood[i];
 

  tmpp3 = tmpp4 = vec;
  *tmpp3++ = *tmpp2++;
  for(j = 1; j < NUM_NUC; ++j)
    *tmpp3++ = *tmpp2++ + *tmpp4++;
  *pc = get_state(NUM_NUC, vec);

  if(*pc != *pc_old)
    changes++;

  for(++i; i < params->tree_time; ++i) {
    if(params->configuration_map[i].type == AXIL) {
      branch_state = params->branch_state[configp->pos] + params->configuration_map[i].node->num;
      if(!branch_state->present) {
	i += params->configuration_map[i].node->time_to_parent - 1;
	continue;
      }
      else if(branch_state->present == 2) {
	i += params->configuration_map[i].node->time_to_parent - 2;
	continue;
      }
      im1 = params->configuration_map[i].node->parent->configuration_index + params->configuration_map[i].node->parent->time_to_parent - 1;
      branch_state_im1 = params->branch_state[configp->pos] + params->configuration_map[im1].node->num;

      pcm1 = params->configurations[branch_state->neighbors ? branch_state->neighbors->left : (configp->pos - 1)] + i;
      pcp1 = params->configurations[branch_state->neighbors ? branch_state->neighbors->right : (configp->pos + 1)] + i;
      
      pcm1_im1 = params->configurations[branch_state_im1->neighbors ? branch_state_im1->neighbors->node_left : (configp->pos - 1)] + im1;
      pcp1_im1 = params->configurations[branch_state_im1->neighbors ? branch_state_im1->neighbors->node_right : (configp->pos + 1)] + im1;
      
      pc = configp->configuration + i;
      pc_old = configp->old_config + i;
      pc_im1 = configp->configuration + im1;
      pc_old_im1 = configp->old_config + im1;
      
      param_table = ((branch_state->direction >= 0) ? params->branch_param_table[hidden_state][branch_state->rate_category] : params->reverse_branch_param_table[hidden_state][branch_state->rate_category]) + params->configuration_map[i].node->num;
    }
    else if(params->configuration_map[i].type == NODE) {
      if(!branch_state->present)
	continue;
      param_table || die("param_table not properly defined: i=%d, pos=%d\n", i, configp->pos);

      pcm1_im1 = pcm1;
      pcp1_im1 = pcp1;
      pc_im1 = pc;
      pc_old_im1 = pc_old;
      
      ++pc;
      ++pc_old;
      
      pcm1 = params->configurations[branch_state->neighbors ? branch_state->neighbors->node_left : (configp->pos - 1)] + i;
      pcp1 = params->configurations[branch_state->neighbors ? branch_state->neighbors->node_right : (configp->pos + 1)] + i;
    }
    else {
      param_table || die("param_table not properly defined y\n");

      pcm1_im1 = pcm1;
      pcp1_im1 = pcp1;
      
      pc_im1 = pc;
      pc_old_im1 = pc_old;
      
      ++pcm1;
      ++pcp1;
      ++pc;
      ++pc_old;
    }
    tmpp = param_table->el[*pcm1_im1][*pc_im1][*pcp1_im1];
    tmpp2 = likelihood[i];
    tmpp3 = tmpp4 = vec;
    *tmpp3++ = *tmpp++ * *tmpp2++;
    for(j = 1; j < NUM_NUC; ++j)
      *tmpp3++ = *tmpp++ * *tmpp2++ + *tmpp4++;
    *pc = get_state(NUM_NUC, vec);
    if(*pc != *pc_old)
      changes++;
  }
  return changes;
}

void conditional_likelihood(Params* params, ConfigP* configp, FLOAT** likelihood) {
  int i, j, k, n;
  FLOAT sum, tmp, tmp2;
  Node* node, *child;
  FLOAT *tmpp, *tmpp2, *tmpp3;
  int seq;
  int first_branch;
  int ip1;
  BranchState* branch_state, *branch_state_ip1, *branch_state_left, *branch_state_right;
  char* pcm1, *pcm2, *pcp1, *pcp2;
  char* pcm1_p1, *pcp1_p1;
  ParamTable* param_table, *param_table_left, *param_table_right;
  RootTable* root_table;
  int left, right, right2;
  FLOAT *tmpl;
  int hidden_state;

  hidden_state = params->hidden_state[configp->pos];

  for(i = 0; i < params->tree_time; ++i)
    for(j = 0; j < NUM_NUC; ++j)
      likelihood[i][j] = 0.0;

  first_branch = 1;
  for(i = params->tree_time - 1; i >= params->root_position[configp->pos]; --i) {
    branch_state = params->branch_state[configp->pos] + params->configuration_map[i].node->num;
    if((seq = params->configuration_map[i].type) >= 0) {
      first_branch = 1;
      if(!branch_state->present) {
	i -= params->configuration_map[i].node->time_to_parent - 1;
	continue;
      }
      if(params->sequence[seq].seq[configp->pos] < NUM_NUC)
	likelihood[i][params->sequence[seq].seq[configp->pos]] = 1.0;
      else
	for(j = 0; j < NUM_NUC; ++j)
	  likelihood[i][j] = 0.25;
    }
    else if(params->configuration_map[i].type != NODE) {
      node = params->configuration_map[i].node;
      if(branch_state->present != 1) {
	i = node->configuration_index;
	continue;
      }
      if(first_branch) {
	branch_state = params->branch_state[configp->pos] + node->num;
	
	left = branch_state->neighbors ? branch_state->neighbors->left : (configp->pos - 1);
	right = branch_state->neighbors ? branch_state->neighbors->right : (configp->pos + 1);


	branch_state_left = params->branch_state[left] + node->num;
	branch_state_right = params->branch_state[right] + node->num;

	pcm1 = params->configurations[left] + i;
	pcp1 = params->configurations[right] + i;
	pcm2 = params->configurations[branch_state_left->neighbors ? branch_state_left->neighbors->left : (left - 1)] + i;
	pcp2 = params->configurations[branch_state_right->neighbors ? branch_state_right->neighbors->right : (right + 1)] + i;
	
	pcm1_p1 = pcm1 + 1;
	pcp1_p1 = pcp1 + 1;

	param_table = ((branch_state->direction >= 0) ? params->branch_param_table[hidden_state][branch_state->rate_category] : params->reverse_branch_param_table[hidden_state][branch_state->rate_category]) + node->num;
	param_table_left = ((branch_state_left->direction >= 0) ? params->branch_param_table[params->hidden_state[left]][branch_state_left->rate_category] : params->reverse_branch_param_table[params->hidden_state[left]][branch_state_left->rate_category]) + node->num;
	param_table_right = ((branch_state_right->direction >= 0) ? params->branch_param_table[params->hidden_state[right]][branch_state_right->rate_category] : params->reverse_branch_param_table[params->hidden_state[right]][branch_state_right->rate_category]) + node->num;

	first_branch = 0;
      }
      else {
	pcm1_p1 = pcm1;
	pcp1_p1 = pcp1;
	--pcm1;
	--pcp1;
	--pcm2;
	--pcp2;
      }
      sum = 0.0; 
      for(j = 0; j < NUM_NUC; j++) {
	tmpp = param_table->el[*pcm1][j][*pcp1];
	tmpp2 = likelihood[i+1];
	tmpp3 = likelihood[i] + j;
	for(k = 0; k < NUM_NUC; ++k)
	  *tmpp3 += *tmpp++ * *tmpp2++;
	*tmpp3 *= param_table_left->el[*pcm2][*pcm1][j][*pcm1_p1] * param_table_right->el[j][*pcp1][*pcp2][*pcp1_p1];
	sum += *tmpp3;
      }
      tmpp3 = likelihood[i];
      for(j = 0; j < NUM_NUC; ++j)
	*tmpp3++ /= sum;
    }
    else { /* == NODE */
      node = params->configuration_map[i].node;
      first_branch = 1;
      if(!branch_state->present) {
	i = node->configuration_index;
	continue;
      }

      branch_state = params->branch_state[configp->pos] + node->num;

      left = branch_state->neighbors ? branch_state->neighbors->node_left : (configp->pos - 1);
      right = branch_state->neighbors ? branch_state->neighbors->node_right : (configp->pos + 1);
      
      pcm1 = params->configurations[left] + i;
      pcp1 = params->configurations[right] + i;

      branch_state_left = params->branch_state[left] + node->num;
      branch_state_right = params->branch_state[right] + node->num;
      pcm2 = params->configurations[branch_state_left->neighbors ? branch_state_left->neighbors->node_left : (left - 1)] + i;
      pcp2 = params->configurations[branch_state_right->neighbors ? branch_state_right->neighbors->node_right : (right + 1)] + i;

      sum = 0.0;
      for(j = 0; j < NUM_NUC; j++) {
	tmpl = likelihood[i] + j;
	*tmpl = 1.0;
	for(n = 0; n < node->n_children; ++n) {
	  child = node->children[n];
	  ip1 = child->configuration_index;
	  branch_state_ip1 = params->branch_state[configp->pos] + child->num;
	  /*
	  if(branch_state_ip1->present != 1)
	    continue;
	  */
	  branch_state_left = params->branch_state[left] + child->num;
	  branch_state_right = params->branch_state[right] + child->num;

	  if(branch_state_ip1->present == 1) {
	    tmp = 0.0;
	    param_table = ((branch_state_ip1->direction >= 0) ? params->branch_param_table[hidden_state][branch_state_ip1->rate_category] : params->reverse_branch_param_table[hidden_state][branch_state_ip1->rate_category]) + child->num;
	    tmpp = param_table->el[*pcm1][j][*pcp1];
	    tmpp2 = likelihood[ip1];
	    for(k = 0; k < NUM_NUC; ++k)
	      tmp += *tmpp++ * *tmpp2++;
	    
	    *tmpl *= tmp;
	  }

	  if(branch_state_left->present == 1) {
	    pcm1_p1 = params->configurations[left] + ip1;
	    param_table_left = ((branch_state_left->direction >= 0) ? params->branch_param_table[params->hidden_state[left]][branch_state_left->rate_category] : params->reverse_branch_param_table[params->hidden_state[left]][branch_state_left->rate_category]) + child->num;	    
	    *tmpl *= param_table_left->el[*pcm2][*pcm1][j][*pcm1_p1];
	  }
	  if(branch_state_right->present == 1) {
	    pcp1_p1 = params->configurations[right] + ip1;
	    param_table_right = ((branch_state_right->direction >= 0) ? params->branch_param_table[params->hidden_state[right]][branch_state_right->rate_category] : params->reverse_branch_param_table[params->hidden_state[right]][branch_state_right->rate_category]) + child->num;
	    *tmpl *= param_table_right->el[j][*pcp1][*pcp2][*pcp1_p1];
	  }
	}
	sum += *tmpl;
      }

      if(i == params->root_position[right]) {
	branch_state_right = params->branch_state[right] + node->num;
	root_table = branch_state_right->direction >= 0 ? &params->cond_root[params->root_category[right]] : &params->reverse_cond_root[params->root_category[right]];
	tmpp3 = likelihood[i];
	for(j = 0; j < NUM_NUC; ++j)
	  *tmpp3++ *= root_table->el[*pcm1][j][*pcp1];
      }
      if(i == params->root_position[configp->pos]) {
	root_table = branch_state->direction >= 0 ? &params->cond_root[params->root_category[configp->pos]] : &params->reverse_cond_root[params->root_category[configp->pos]];
	tmpp3 = likelihood[i];
	for(j = 0; j < NUM_NUC; ++j)
	  *tmpp3++ *= root_table->el[*pcm2][*pcm1][j];
      }
      branch_state_right = params->branch_state[right] + node->num;
      right2 = branch_state_right->neighbors ? branch_state_right->neighbors->node_right : (right + 1);
      if(i == params->root_position[right2]) {
	branch_state_right = params->branch_state[right2] + node->num;
	root_table = branch_state_right->direction >= 0 ? &params->cond_root[params->root_category[right2]] : &params->reverse_cond_root[params->root_category[right2]];
	tmpp3 = likelihood[i];
	for(j = 0; j < NUM_NUC; ++j)
	  *tmpp3++ *= root_table->el[j][*pcp1][*pcp2];
      }


      tmpp3 = likelihood[i];
      for(j = 0; j < NUM_NUC; ++j)
	*tmpp3++ /= sum;
    }
  }

  /*
  i = params->root_position[configp->pos];
  node = params->configuration_map[i].node;
  branch_state = params->branch_state[configp->pos] + node->num;

  left = branch_state->neighbors ? branch_state->neighbors->node_left : (configp->pos - 1);
  right = branch_state->neighbors ? branch_state->neighbors->node_right : (configp->pos + 1);

  branch_state_left = params->branch_state[left] + node->num;
  branch_state_right = params->branch_state[right] + node->num;

  pcm1 = params->configurations[left] + i;
  pcp1 = params->configurations[right] + i;
  pcm2 = params->configurations[branch_state_left->neighbors ? branch_state_left->neighbors->node_left : (left - 1)] + i;
  pcp2 = params->configurations[branch_state_right->neighbors ? branch_state_right->neighbors->node_right : (right + 1)] + i;


  root_table = branch_state->direction >= 0 ? &params->cond_root[params->root_category[configp->pos]] : &params->reverse_cond_root[params->root_category[configp->pos]];
  root_table_left = branch_state_left->direction >= 0 ? &params->cond_root[params->root_category[left]] : &params->reverse_cond_root[params->root_category[left]];
  root_table_right = branch_state_right->direction >= 0 ? &params->cond_root[params->root_category[right]] : &params->reverse_cond_root[params->root_category[right]];


  tmpp3 = likelihood[i];
  if(params->root_position[left] == i) {
    if(params->root_position[right] == i)
      for(j = 0; j < NUM_NUC; ++j)
	*tmpp3++ *= root_table_left->el[*pcm2][*pcm1][j] * root_table->el[*pcm1][j][*pcp1] * root_table_right->el[j][*pcp1][*pcp2];
    else
      for(j = 0; j < NUM_NUC; ++j)
	*tmpp3++ *= root_table_left->el[*pcm2][*pcm1][j] * root_table->el[*pcm1][j][*pcp1];
  }
  else {
    if(params->root_position[right] == i)
      for(j = 0; j < NUM_NUC; ++j)
	*tmpp3++ *= root_table->el[*pcm1][j][*pcp1] * root_table_right->el[j][*pcp1][*pcp2];
    else
      for(j = 0; j < NUM_NUC; ++j)
	*tmpp3++ *= root_table->el[*pcm1][j][*pcp1];
  }
  */
}

void iterate_variation(Params* params) {
  int i;

  if(params->variation) {
    fprintf(stderr, "a");
    update_branch_param_list(params);
    emission_probabilities(params);

    for(i = 0; i < 8; i++) {
      fprintf(stderr, "b");
      backward_variables(params);
      fprintf(stderr, "c");
      sample_hidden_states(params);
      fprintf(stderr, "d");
      if(!params->fix_transition_parameters)
	iterate_variation_parameters(params);
      else
	break;
    }
    count_configuration_changes(params);
    fprintf(stderr, "e");
  }
}

void iterate_variation_parameters(Params* params) {
  int* count, **dicount;
  int i, h;

  (count = (int*)malloc(sizeof(int) * params->num_hidden_states)) || die("Could not allocate memory for count\n");
  (dicount = (int**)malloc(sizeof(int*) * params->num_hidden_states)) || die("Could not allocate memory for dicount\n");
  for(i = 0; i < params->num_hidden_states; i++) {
    count[i] = 0;
    (dicount[i] = (int*)malloc(sizeof(int) * params->num_hidden_states)) || die("Could not allocate memory for dicount[%d]\n", i);
    for(h = 0; h < params->num_hidden_states; h++)
      dicount[i][h] = 0;
  }

  ++count[params->hidden_state[0]];

  for(i = 1; i < params->seq_length; i++) {
    ++count[params->hidden_state[i]];
    ++dicount[params->hidden_state[i-1]][params->hidden_state[i]];
  }
  /*
  sample_from_counts(params->num_hidden_states, count, params->initial);
  */
  for(i = 0; i < params->num_hidden_states; i++)
    sample_from_counts(params->last_unfixed_transition[i], dicount[i], params->transition[i], params->fix_transition[i]);
  
  for(i = 0; i < params->num_hidden_states; i++)
    free(dicount[i]);
  free(dicount);
  free(count);

  compute_stationary_from_transition(params);
}

void sample_from_counts(int n, int* count, double* parameters, int* fixed) {
  double* new;
  double try;
  int sum;
  int i, s, j;
  double ratio;
  double x, y, z;
  double mean, variance;

  if(n < 2)
    return;

  (new = (double*)malloc(sizeof(double) * n)) || die("Could not allocate memory for new\n");

  sum = 0;
  for(i = 0; i < n; i++)
    sum += count[i];

  for(s = 0; s < 8; s++)
    for(i = 0; i < n - 1; i++) 
      if(!fixed[i]) {
	/*
	  for(j = 0; j < n; j++)
	  fprintf(stderr, "%d %g ", count[j], parameters[j]);
	  fprintf(stderr, "\n");
	*/
	mean = (double)count[i] / sum;
	variance = (double)mean / sum;
	try = random_t(mean, variance, 4);
	/*
	  fprintf(stderr, "mean = %g, variance = %g, new[%d] = %g, old = %g, ratio = %g\n", mean, variance, i, try, parameters[i], ratio);
	*/
	if(try > 0.0 && try < 1.0 && try - parameters[i] < parameters[n-1]) {
	  x = parameters[i] - mean;
	  y = try - mean;
	  z = 4.0 * variance;
	  ratio = pow((z + y * y) / (z + x * x), 2.5) * pow(try / parameters[i], count[i]) * pow((parameters[n-1] - try + parameters[i]) / parameters[n-1], count[n-1]);
	  if(ratio >= 1.0 || random_double(0.0, 1.0) < ratio) {
	    parameters[n-1] += parameters[i] - try;
	    parameters[i] = try;
	  }
	}
      }
  
  free(new);
}

void initiate_configuration(Params* params) {
  int i, j;


  for(i = 0; i < params->seq_length; i++)
    for(j = 0; j < params->tree_time; j++)
      if(params->configuration_map[j].type >= 0 && params->sequence[params->configuration_map[j].type].seq[i] < NUM_NUC)
	params->configurations[i][j] = params->sequence[params->configuration_map[j].type].seq[i];
      else
	params->configurations[i][j] = (char)random_int(0, NUM_NUC - 1);
}

void initiate_variation(Params* params) {
  int i, j, last;
  double *vec;

  if(params->variation) {
    (vec = (double*)malloc(sizeof(double) * params->num_hidden_states)) || die("Could not allocate memory for vec\n");
    for(i = 0; i < params->num_hidden_states; i++)
      vec[i] = params->initial[i] + (i ? vec[i-1] : 0.0);
    last = params->hidden_state[0] = get_state(params->num_hidden_states, vec);

    for(i = 1; i < params->seq_length; i++) {
      for(j = 0; j < params->num_hidden_states; j++)
	vec[j] = params->transition[last][j] + (j ? vec[j-1] : 0.0);
      last = params->hidden_state[i] = get_state(params->num_hidden_states, vec);
    }
    free(vec);
  }
  else
    for(i = 0; i < params->seq_length; i++)
      params->hidden_state[i] = 0;
}

int get_state(int n, double* vector) {
  double r;
  int high, low, test;

  if(n == 1)
    return 0;

  r = random_double(0.0, vector[n-1]);

  low = 0;
  high = n - 1;
  while (1) {
    if(low == high)
      return low;
    test = (high + low) / 2;
    if(r < vector[test])
      high = test;
    else
      low = test + 1;
  }
}

void emission_probabilities(Params* params) {
  int i, b, j, h;
  PositionChanges** changes;
  double probability;
  BranchState* branch_state;
  double max;
  

  (changes = (PositionChanges**)malloc(sizeof(PositionChanges*) * params->num_branches)) || die("Could not allocate memory for changes\n");
  for(i = 0; i < params->num_branches; i++)
    (changes[i] = (PositionChanges*)malloc(sizeof(PositionChanges) * (params->branch_length[i] + 1))) || die("Could not allocate memory for changes[%d]\n", i);
  

  for(i = 0; i < params->seq_length; i++) {
    count_position_changes(params, changes, i);
    for(h = 0; h < params->num_hidden_states; h++) {
      probability = 0.0;
      for(branch_state = params->branch_state[i], b = 0; b < params->num_branches; b++, branch_state++)
	for(j = 0; changes[b][j].num; j++)
	  probability += changes[b][j].num * log(params->branch_param_list[h][branch_state->rate_category][b][changes[b][j].change]);
      params->emission[i][h] = probability;
      if(!h || probability > max)
	max = probability;
      
    }
    for(h = 0; h < params->num_hidden_states; h++)
      params->emission[i][h] = exp(params->emission[i][h] - max);
  }
  
  for(i = 0; i < params->num_branches; i++)
    free(changes[i]);
  free(changes);
}

/*
void forward_variables(Params* params) {
  int i, h, j;
  double sum, **forward, **forwardm1, **emission;

  for(h = 0; h < params->num_hidden_states; h++)
    params->forward[0][h] = params->emission[0][h] * params->initial[h];

  for(emission = params->emission, forwardm1 = params->forward, forward = params->forward + 1, i = 1; i < params->seq_length; i++, forward++, forwardm1++, emission++) {
    sum = 0.0;
    for(h = 0; h < params->num_hidden_states; h++) {
      *forward[h] = 0.0;
      for(j = 0; j < params->num_hidden_states; j++)
	*forward[h] += *forwardm1[j] * params->transition[j][h];
      *forward[h] *= *emission[h];
      sum += *forward[h];
    }
    for(h = 0; h < params->num_hidden_states; h++)
      *forward[h] /= sum;
  }
}
*/

void backward_variables(Params* params) {
  int i, h, j;
  double sum;

  i = params->seq_length - 1;
  for(h = 0; h < params->num_hidden_states; h++) 
    params->backward[i][h] = 1.0;
  
  for(i-- ; i >= 0; i--) {
    sum = 0.0;
    for(h = 0; h < params->num_hidden_states; h++) {
      params->backward[i][h] = 0.0;
      for(j = 0; j < params->num_hidden_states; j++)
	params->backward[i][h] += params->transition[h][j] * params->backward[i+1][j] * params->emission[i+1][j];
      sum += params->backward[i][h];
    }
    for(h = 0; h < params->num_hidden_states; h++)
      params->backward[i][h] /= sum;
  }
}

void sample_hidden_states(Params* params) {
  int i, h;
  int last;
  double* vec;

  (vec = (double*)malloc(sizeof(double) * params->num_hidden_states)) || die("Could not allocate memory for vec\n");
  
  vec[0] = params->initial[0] * params->emission[0][0] * params->backward[0][0];
  for(h = 1; h < params->num_hidden_states; h++)
    vec[h] = vec[h-1] + params->initial[h] * params->emission[0][h] * params->backward[0][h];
  last = params->hidden_state[0] = get_state(params->num_hidden_states, vec);
  if(params->record)
    params->variation_count[0][last]++;

  for(i = 1; i < params->seq_length; i++) {
    vec[0] = params->transition[last][0] * params->emission[i][0] * params->backward[i][0];
    for(h = 1; h < params->num_hidden_states; h++)
      vec[h] = vec[h-1] + params->transition[last][h] * params->emission[i][h] * params->backward[i][h];
    last = params->hidden_state[i] = get_state(params->num_hidden_states, vec);
    if(params->record)
      params->variation_count[i][last]++;
  }

  free(vec);
}

void count_position_changes(Params* params, PositionChanges** position_changes, int pos) {
  int i, j, jm1;
  char* pcm1, *pcp1;
  char* pcm1_im1, *pcp1_im1;
  char* pc, *pc_im1;
  BranchState* branch_state, *branch_state_im1, *branch_state_left;
  int* trinuc_parameters;
  PositionChanges *changes, *this;
  Node* node;
  int axil2;
  int left, right, left2;
  int change;

  for(i = 0; i < params->num_branches; i++)
    position_changes[i][0].num = 0;
  
  j = params->root_position[pos];
  if(j == -1)
    return;
  branch_state = params->branch_state[pos] + params->configuration_map[j].node->num;
  left = branch_state->neighbors ? branch_state->neighbors->node_left : (pos-1);
  branch_state_left = params->branch_state[left] + params->configuration_map[j].node->num;
  left2 = branch_state_left->neighbors ? branch_state_left->neighbors->node_left : (left - 1);
  changes = NULL;
  for(j++; j < params->tree_time; j++) {
    if(params->configuration_map[j].type == AXIL) {
      axil2 = 1;
      node = params->configuration_map[j].node;
      branch_state = params->branch_state[pos] + node->num;
      
      if(branch_state->present != 1) {
	j += node->time_to_parent - 1;
	continue;
      }
      changes = position_changes[node->num];
      trinuc_parameters = (branch_state->direction >= 0) ? params->trinuc_parameters[branch_state->direction] : params->complement_trinuc_parameters;
      jm1 = node->parent->configuration_index + node->parent->time_to_parent - 1;
      branch_state_im1 = params->branch_state[pos] + node->parent->num;
      
      left = branch_state_im1->neighbors ? branch_state_im1->neighbors->node_left : (pos-1);
      right = branch_state_im1->neighbors ? branch_state_im1->neighbors->node_right : (pos+1);
      
      pcm1_im1 = params->configurations[left] + jm1;
      pcp1_im1 = params->configurations[right] + jm1;
      
      pc = params->configurations[pos] + j;
      pc_im1 = params->configurations[pos] + jm1;
      
    }
    else {
      changes || die("changes not properly defined\n");
      
      if(axil2) {
	left = branch_state->neighbors ? branch_state->neighbors->left : (pos - 1);
	right = branch_state->neighbors ? branch_state->neighbors->right : (pos + 1);
	
	pcm1 = params->configurations[left] + j;
	pcp1 = params->configurations[right] + j;
	
	pcm1_im1 = pcm1 - 1;
	pcp1_im1 = pcp1 - 1;
	
	pc_im1 = pc;
	++pc;
	
	axil2 = 0;
      }
      else {
	pcm1_im1 = pcm1;
	pcp1_im1 = pcp1;
	
	pc_im1 = pc;
	
	++pcm1;
	++pcp1;
	++pc;
      }
    }
    change = trinuc_parameters[64 * *pcm1_im1 + 16 * *pc_im1 + 4 * *pcp1_im1 + *pc];
    
    for(this = changes; ; this++) {
      if(this->num == 0) {
	this->change = change;
	this->num = 1;
	this[1].num = 0;
	break;
      }
      else if(change == this->change) {
	++this->num;
	break;
      }
    }
  }
}

void count_configuration_changes(Params* params) {
  int b, d, i, j, jm1;
  char* pcm1, *pcp1;
  char* pcm1_im1, *pcp1_im1;
  char* pc, *pc_im1;
  BranchState* branch_state, *branch_state_im1, *branch_state_left;
  int* trinuc_parameters;
  int* changes;
  Node* node;
  int axil2;
  int left, right, left2;
  int hidden_state;
  
  for(j = 0; j < params->num_hidden_states; j++)
    for(d = 0; d < params->num_rate_categories; d++)
      for(b = 0; b < params->num_branches; b++)
	for(i = 0; i < params->num_total_mutation_parameters[params->category_type[d]]; i++)
	  params->changes[j][d][b][i] = 0;
  for(d = 0; d < params->num_root_categories; d++)
    for(i = 0; i < 64; i++)
      params->root_changes[d][i] = 0;
  
  for(i = 0; i < params->seq_length; i++) {
    hidden_state = params->hidden_state[i];
    j = params->root_position[i];
    if(j == -1)
      continue;
    branch_state = params->branch_state[i] + params->configuration_map[j].node->num;
    left = branch_state->neighbors ? branch_state->neighbors->node_left : (i-1);
    branch_state_left = params->branch_state[left] + params->configuration_map[j].node->num;
    left2 = branch_state_left->neighbors ? branch_state_left->neighbors->node_left : (left - 1);
    if(branch_state->direction >= 0)
      ++params->root_changes[params->root_category[i]][params->root_parameters[params->configurations[left2][j]][params->configurations[left][j]][params->configurations[i][j]]];
    else
      ++params->root_changes[params->root_category[i]][params->complement_root_parameters[params->configurations[left2][j]][params->configurations[left][j]][params->configurations[i][j]]];
    changes = NULL;
    for(j++; j < params->tree_time; j++) {
      if(params->configuration_map[j].type == AXIL) {
	axil2 = 1;
	node = params->configuration_map[j].node;
	branch_state = params->branch_state[i] + node->num;

	if(branch_state->present != 1) {
	  j += node->time_to_parent - 1;
	  continue;
	}
	changes = params->changes[hidden_state][branch_state->rate_category][node->num];
	trinuc_parameters = (branch_state->direction >= 0) ? params->trinuc_parameters[branch_state->direction] : params->complement_trinuc_parameters;
	jm1 = node->parent->configuration_index + node->parent->time_to_parent - 1;
	branch_state_im1 = params->branch_state[i] + node->parent->num;

	left = branch_state_im1->neighbors ? branch_state_im1->neighbors->node_left : (i-1);
	right = branch_state_im1->neighbors ? branch_state_im1->neighbors->node_right : (i+1);
	
	pcm1_im1 = params->configurations[left] + jm1;
	pcp1_im1 = params->configurations[right] + jm1;
	
	pc = params->configurations[i] + j;
	pc_im1 = params->configurations[i] + jm1;
	
      }
      else {
	changes || die("changes not properly defined\n");

	if(axil2) {
	  left = branch_state->neighbors ? branch_state->neighbors->left : (i - 1);
	  right = branch_state->neighbors ? branch_state->neighbors->right : (i + 1);

	  pcm1 = params->configurations[left] + j;
	  pcp1 = params->configurations[right] + j;
	  
	  pcm1_im1 = pcm1 - 1;
	  pcp1_im1 = pcp1 - 1;
	
	  pc_im1 = pc;
	  ++pc;

	  axil2 = 0;
	}
	else {
	  
	  pcm1_im1 = pcm1;
	  pcp1_im1 = pcp1;
	  
	  pc_im1 = pc;
	  
	  ++pcm1;
	  ++pcp1;
	  ++pc;
	}
      }
      ++changes[trinuc_parameters[64 * *pcm1_im1 + 16 * *pc_im1 + 4 * *pcp1_im1 + *pc]];
    }
  }
}

void enstate_branches(Params* params) {
  int i, j, count;
  Neighbors* neighbors;

  (params->root_position = (short*)malloc(sizeof(short) * params->seq_length)) || die("Could not allocate memory for params->root_position\n");
  (params->branch_state = (BranchState**)malloc(sizeof(BranchState*) * params->seq_length)) || die("Could not allocate memory for params->branch_state\n");

  for(i = 0; i < params->seq_length; i++) {
    (params->branch_state[i] = (BranchState*)malloc(sizeof(BranchState) * (params->num_branches + 1))) || die("Could not allocate memory for params->branch_state[%d]\n", i);
    for(j = 0; j <= params->num_branches; j++)
      params->branch_state[i][j].neighbors = NULL;
    params->root_position[i] = -1;
  }

  instate_node1(params, params->root);
  instate_node2(params, params->root);

  count = 0;
  for(i = 0; i < params->seq_length; i++)
    for(j = 0; j <= params->num_branches; j++)
      if(params->branch_state[i][j].neighbors)
	count++;
  
  (neighbors = (Neighbors*)malloc(sizeof(Neighbors) * count)) || die("Could not allocate memory for neighbors\n");
  fprintf(stderr, "Counted %d non-standard neighbors\n", count);
  
  for(i = 0; i < params->seq_length; i++)
    for(j = 0; j <= params->num_branches; j++)
      if(params->branch_state[i][j].neighbors)
	params->branch_state[i][j].neighbors = neighbors++;

  instate_node3(params, params->root);
}

void instate_node1(Params* params, Node* node) {
  int i, n;
  char* seq;

  if(!node->n_children) {
    seq = params->sequence[node->seq_num].seq;
    for(i = 0; i < params->seq_length; i++)
      params->branch_state[i][node->num].present = (*seq++ <= NUM_NUC);
  }
  else {
    for(n = 0; n < node->n_children; n++)
      instate_node1(params, node->children[n]);
    for(i = 0; i < params->seq_length; i++) {
      params->branch_state[i][node->num].present = 0;
      for(n = 0; n < node->n_children; n++)
	if(params->branch_state[i][node->children[n]->num].present)
	  params->branch_state[i][node->num].present++;
    }
  }
}

void instate_node2(Params* params, Node* node) {
  int i, n, last, first, node_last, node_first;
  
  last = node_last = 0;
  first = node_first = -1;
  for(i = 0; i < params->seq_length; i++) {
    if(params->branch_state[i][node->num].present > 1 || params->branch_state[i][node->num].present && node->parent && params->branch_state[i][node->parent->num].present || params->branch_state[i][node->num].present && !node->n_children) {
      if(!node->parent || !params->branch_state[i][node->parent->num].present) {
	params->branch_state[i][node->num].present = 2;
	params->root_position[i] = node->configuration_index + node->time_to_parent - 1;
      }
      else {
	params->branch_state[i][node->num].present = 1;
	if(last != i - 1)
	  params->branch_state[i][node->num].neighbors = (Neighbors*)1;
	last = i;
	if(first == -1)
	  first = i;
      }
      if(node_last != i - 1)
	params->branch_state[i][node->num].neighbors = (Neighbors*)1;
      node_last = i;
      if(node_first == -1)
	node_first = i;
    }
    else
      params->branch_state[i][node->num].present = 0;
  }
  if(first != -1 && last != first - 1)
    params->branch_state[first][node->num].neighbors = (Neighbors*)1;
  
  if(node_first != -1 && node_last != node_first - 1)
    params->branch_state[node_first][node->num].neighbors = (Neighbors*)1;
  
  for(i = params->seq_length - 1; i >= 0; i--)
    if(params->branch_state[i][node->num].present) {
      if(node_first != i + 1)
	params->branch_state[i][node->num].neighbors = (Neighbors*)1;
      node_first = i;
      if(params->branch_state[i][node->num].present == 1) {
	if(first != i + 1)
	  params->branch_state[i][node->num].neighbors = (Neighbors*)1;
	first = i;
      }
    }
  for(n = 0; n < node->n_children; n++)
    instate_node2(params, node->children[n]);
}

void instate_node3(Params* params, Node* node) {
  int i, n, last, first, node_last, node_first;
  
  last = node_last = 0;
  first = node_first = -1;
  for(i = 0; i < params->seq_length; i++) {
    if(params->branch_state[i][node->num].present > 1 || params->branch_state[i][node->num].present && node->parent && params->branch_state[i][node->parent->num].present || params->branch_state[i][node->num].present && !node->n_children) {
      if(!node->parent || !params->branch_state[i][node->parent->num].present) {
	params->branch_state[i][node->num].present = 2;
	params->root_position[i] = node->configuration_index + node->time_to_parent - 1;
      }
      else {
	params->branch_state[i][node->num].present = 1;
	if(params->branch_state[i][node->num].neighbors)
	  params->branch_state[i][node->num].neighbors->left = last;
	last = i;
	if(first == -1)
	  first = i;
      }
      if(params->branch_state[i][node->num].neighbors)
	params->branch_state[i][node->num].neighbors->node_left = node_last;
      node_last = i;
      if(node_first == -1)
	node_first = i;
    }
    else
      params->branch_state[i][node->num].present = 0;
  }
  if(first != -1 && params->branch_state[first][node->num].neighbors) 
    params->branch_state[first][node->num].neighbors->left = last;
  if(node_first != -1 && params->branch_state[node_first][node->num].neighbors)
    params->branch_state[node_first][node->num].neighbors->node_left = node_last;
  for(i = params->seq_length - 1; i >= 0; i--)
    if(params->branch_state[i][node->num].present) {
      if(params->branch_state[i][node->num].neighbors)
	params->branch_state[i][node->num].neighbors->node_right = node_first;
      node_first = i;
      if(params->branch_state[i][node->num].present == 1) {
	if(params->branch_state[i][node->num].neighbors)
	  params->branch_state[i][node->num].neighbors->right = first;
	first = i;
      }
    }
  for(n = 0; n < node->n_children; n++)
    instate_node3(params, node->children[n]);
}
 
void create_configuration(Params* params) {
  params->tree_time = configuration_index(params);
  (params->configuration_map = (Map*)malloc(sizeof(Map) * params->tree_time)) || die("Could not allocate memory for params->configuration_index\n");
  fill_configuration_map(params, params->root);
}

void allocate_configuration(Params* params) {
  int i;

  (params->configurations = (char**)malloc(sizeof(char*) * params->seq_length)) || die("Could not allocate memory for params->configurations\n");
  for(i = 0; i < params->seq_length; i++) 
    (params->configurations[i] = (char*)malloc(sizeof(char) * params->tree_time)) || die("Could not allocate memory for params->configurations[%d]\n", i);
}

void allocate_variation(Params* params) {
  int i, j;

  (params->hidden_state = (unsigned char*)malloc(sizeof(unsigned char) * params->seq_length)) || die("Could not allocate memory for params->hidden_state\n");
  (params->emission = (double**)malloc(sizeof(double*) * params->seq_length)) || die("Could not allocate memory for params->emission\n");
  (params->backward = (double**)malloc(sizeof(double*) * params->seq_length)) || die("Could not allocate memory for params->backward\n");
  for(i = 0; i < params->seq_length; i++) {
    (params->emission[i] = (double*)malloc(sizeof(double) * params->num_hidden_states)) || die("Could not allocate memory for params->emission[%d]\n", i);
    (params->backward[i] = (double*)malloc(sizeof(double) * params->num_hidden_states)) || die("Could not allocate memory for params->backward[%d]\n", i);
  }
  (params->initial = (double*)malloc(sizeof(double) * params->num_hidden_states)) || die("Could not allocate memory for params->initial\n");

  if(!params->variation) {
    (params->rate_multiplier = (double*)malloc(sizeof(double) * params->num_hidden_states)) || die("Could not allocate memory for params->rate_multiplier\n");
    (params->transition = (double**)malloc(sizeof(double*) * params->num_hidden_states)) || die("Could not allocate memory for params->transition\n");
    for(i = 0; i < params->num_hidden_states; i++)
      (params->transition[i] = (double*)malloc(sizeof(double) * params->num_hidden_states)) || die("Could not allocate memory for params->transition[%d]\n", i);
    
    params->transition[0][0] = 1.0;
    params->initial[0] = 1.0;
    params->rate_multiplier[0] = 1.0;
  }
}

void compute_stationary_from_transition(Params* params) {
  double** mat1, **mat2, **tmp;
  int i, j, k;
  int good;
  
  (mat1 = (double**)malloc(sizeof(double*) * params->num_hidden_states)) || die("Could not allocate memory for mat1\n");
  (mat2 = (double**)malloc(sizeof(double*) * params->num_hidden_states)) || die("Could not allocate memory for mat2\n");
  for(i = 0; i < params->num_hidden_states; i++) {
    (mat1[i] = (double*)malloc(sizeof(double) * params->num_hidden_states)) || die("Could not allocate memory for mat1[%d]\n", i);
    (mat2[i] = (double*)malloc(sizeof(double) * params->num_hidden_states)) || die("Could not allocate memory for mat2[%d]\n", i);
  }

  for(i = 0; i < params->num_hidden_states; i++)
    for(j = 0; j < params->num_hidden_states; j++)
      mat1[i][j] = params->transition[i][j];

  for(i = 0; i < 256; i++) {
    matrix_multiply(mat2, mat1, mat1, params->num_hidden_states);
    matrix_multiply(mat1, mat2, mat2, params->num_hidden_states);

    good = 1;
    for(j = 0; good && j < params->num_hidden_states; j++)
      for(k = 1; k < params->num_hidden_states; k++)
	if(fabs((mat1[0][j] - mat1[k][j]) / mat1[0][j]) > 1.0e-6) {
	  good = 0;
	  break;
	}
    if(good)
      break;
  }
  if(i == 256)
    die("No convergence in stationary distribution after 2^512 iterations\n");

  /*
  fprintf(stderr, "Converged in stationary after 2^%d iterations\n", 2 * i);
  */

  for(i = 0; i < params->num_hidden_states; i++)
    params->initial[i] = mat1[0][i];

  for(i = 0; i < params->num_hidden_states; i++) {
    free(mat1[i]);
    free(mat2[i]);
  }
  free(mat1);
  free(mat2);
}

void matrix_multiply(double** c, double** a, double** b, int n) {
  int i, j, k;
  double* tmpc, *tmpa;

  for(i = 0; i < n; i++) {
    tmpc = c[i];
    for(j = 0; j < n; j++, tmpc++) {
      tmpa = a[i];
      *tmpc = 0.0;
      for(k = 0; k < n; k++)
	*tmpc += *tmpa++ * b[k][j];
    }
  }
}

void fill_configuration_map(Params* params, Node* node) {
  int i, n;

  if(!node)
    return;

  for(i = node->configuration_index; i < node->configuration_index + node->time_to_parent; i++)
    params->configuration_map[i].node = node;
  for(i = node->configuration_index + 1; i < node->configuration_index + node->time_to_parent - 1; i++)
    params->configuration_map[i].type = BRANCH;
  if(node == params->root)
    params->configuration_map[0].type = NODE;
  else {
    params->configuration_map[node->configuration_index].type = AXIL;
    if(node->n_children)
      params->configuration_map[node->configuration_index + node->time_to_parent - 1].type = NODE;
    else
      params->configuration_map[node->configuration_index + node->time_to_parent - 1].type = (char)node->seq_num;
  }
  
  for(n = 0; n < node->n_children; n++)
    fill_configuration_map(params, node->children[n]);
}

int configuration_index(Params* params) {
  int index = 0;

  node_index(params, params->root, &index);
  return index;
}

void node_index(Params* params, Node* node, int* index) {
  int n;

  if(!node)
    return;
  node->configuration_index = *index;
  *index += node->time_to_parent;
  for(n = 0; n < node->n_children; n++)
    node_index(params, node->children[n], index);
}

void read_params(int argc, char* argv[], Params* params) {
  int i, j;
  char* mapfile;
  char outname[MAX_LINE];
  double sample_rate;
  char* mutation_type_file;
  char* variation_file;

  params->samples = -1;
  sample_rate = DEFAULT_SAMPLE_RATE;
  params->burnin = DEFAULT_BURNIN;
  params->alpha = DEFAULT_ALPHA;
  params->max_rate = DEFAULT_MAX_MUTATION_RATE;
  params->update_parameters_rate = DEFAULT_UPDATE_PARAMETERS_RATE;
  params->update_root_rate = DEFAULT_UPDATE_ROOT_RATE;
  params->seed = 0;
  params->probability = 0;
  mapfile = NULL;
  params->out_prefix = "bayes";
  params->node_out = NULL;
  params->changes_out = NULL;
  params->num_mutation_types = 1;
  params->variation = 0;
  params->fix_transition_parameters = 0;
  params->stats = 0;
    
  for(i = 1; i < argc && argv[i][0] == '-'; i++) {
    if(!strcmp(argv[i] + 1, "rate"))
      sample_rate = atoi(argv[++i]);
    else if(!strcmp(argv[i] + 1, "samples"))
      params->samples = atoi(argv[++i]);
    else if(!strcmp(argv[i] + 1, "seed"))
      params->seed = atoi(argv[++i]);
    else if(!strcmp(argv[i] + 1, "burnin"))
      params->burnin = atoi(argv[++i]);
    else if(!strcmp(argv[i] + 1, "alpha"))
      params->alpha = atof(argv[++i]);
    else if(!strcmp(argv[i] + 1, "max_rate"))
      params->max_rate = atof(argv[++i]);
    else if(!strcmp(argv[i] + 1, "update_parameters_rate"))
      params->update_parameters_rate = atof(argv[++i]);
    else if(!strcmp(argv[i] + 1, "update_root_rate"))
      params->update_root_rate = atof(argv[++i]);
    else if(!strcmp(argv[i] + 1, "map"))
      mapfile = argv[++i];
    else if(!strcmp(argv[i] + 1, "probability"))
      params->probability = 1;
    else if(!strcmp(argv[i] + 1, "out"))
      params->out_prefix = argv[++i];
    else if(!strcmp(argv[i] + 1, "node_count"))
      params->node_out = (FILE*)1;
    else if(!strcmp(argv[i] + 1, "changes_count"))
      params->changes_out = (FILE***)1;
    else if(!strcmp(argv[i] + 1, "mutation_types")) {
      params->num_mutation_types = atoi(argv[++i]);
      if(params->num_mutation_types > 1)
	mutation_type_file = argv[++i];
    }
    else if(!strcmp(argv[i] + 1, "stats"))
      params->stats = 1;
    else
      die(USAGE);
  }

  argc == i + 4 || die(USAGE);
  fprintf(stderr, "Reading parameters...\n");
  read_sequence(params, argv[i++]);
  markov(params);
  if(params->seq_length <= 10000)
    params->sample_rate = params->seq_length;
  else
    params->sample_rate = floor(params->seq_length * sample_rate + 0.5);
  if(params->samples == -1) {
    if(params->seq_length <= 10000)
      params->samples = 1000;
    else
      params->samples = DEFAULT_SAMPLES;
  }    
    
  fprintf(stderr, "Samples recorded every %d missing data updates\n", params->sample_rate);
  make_trinuc_parameters(params, mutation_type_file);

  fprintf(stderr, "Reading tree...\n");
  read_tree(params, argv[i++]);

  read_mapfile(params, mapfile);

  make_root_parameters(params);

  create_configuration(params);

  fprintf(stderr, "Enstating branches...\n");
  enstate_branches(params);

  if(params->variation) 
    read_variation_file(params, variation_file);
  else
    params->num_hidden_states = 1;

  allocate_variation(params);

  fprintf(stderr, "Reading rate categories...\n");
  read_rate_categories(params, argv[i++]);
  fprintf(stderr, "Reading root categories...\n");
  read_root_categories(params, argv[i++]);
    
  (params->branch_out = (FILE**)malloc(sizeof(FILE*) * params->num_mutation_types)) || die("Could not allocate memory for params->branch_out\n");
  for(i = 0; i < params->num_mutation_types; i++) {
    if(params->mutation_types == 1)
      sprintf(outname, "%smcmc.sym.branch", params->out_prefix);
    else
      sprintf(outname, "%smcmc.mutationtype_%.2d.branch", params->out_prefix, i);
    (params->branch_out[i] = fopen(outname, "w")) || die("Could not create file: %s\n", outname);
  }
  if(params->node_out) {
    sprintf(outname, "%smcmc.node_0.sequence", params->out_prefix);
    (params->node_out = fopen(outname, "w")) || die("Could not create file: %s\n", outname);
  }
  if(params->changes_out) {
    (params->changes_out = (FILE***)malloc(sizeof(FILE**) * params->num_rate_categories)) || die("Could not allocate memory for params->changes_out\n");
    for(i = 0; i < params->num_rate_categories; i++) {
      (params->changes_out[i] = (FILE**)malloc(sizeof(FILE*) * params->num_branches)) || die("Could not allocate memory for params->changes_out[%d]\n", i);
      for(j = 0; j < params->num_branches; j++)
	if(params->category_branch_existence[i][j]) {
	  sprintf(outname, "%smcmc.%s.branch_%d.change", params->out_prefix, params->category_name[i], j);
	  (params->changes_out[i][j] = fopen(outname, "w")) || die("Could not create file: %s\n", outname);
	}
    }
  }
    

  allocate_configuration(params);

  if(params->seed)
    srandom(params->seed);
  else
    randomize();
}


void read_mapfile(Params* params, char* mapfile) {
  int i;
  FILE *in;

  if(!mapfile) {
    (params->map = (int*)malloc(sizeof(int) * params->seq_length)) || die("Could not allocate memory for params->map\n");
    params->map_length = params->seq_length;
    for(i = 0; i < params->seq_length; i++)
      params->map[i] = i;
  }
  else {
    (in = fopen(mapfile, "r")) || die("Could not open mapfile: %s\n", mapfile);
    fscanf(in, "%d", &params->map_length) == 1 || die("Error reading length in mapfile: %s\n", mapfile);
    (params->map = (int*)malloc(sizeof(int) * params->map_length)) || die("Could not allocate memory for params->map\n");
    for(i = 0; i < params->map_length; i++) {
      fscanf(in, "%d", params->map + i) == 1 || die("Error reading value %d in mapfile: %s\n", i, mapfile);
      params->map[i] >= -1 && params->map[i] < params->seq_length || die("Map value is out of range: %d\n", params->map[i]);
    }
  }
}

void make_trinuc_parameters(Params* params, char* mutation_type_file) {
  int i, j, k, l, a;
  int count;
  int tmp;
  FILE* in;
  char line[MAX_LINE];
  char c[8];
  int value;
  

  for(a = 0; a < 2; a++) {
    if(a) {
      params->num_mutation_parameters[a] = 192;
      params->num_total_mutation_parameters[a] = 256;
      params->num_mutation_set[a] = 1;
      (params->param_complement = (int*)malloc(sizeof(int) * params->num_total_mutation_parameters[a])) || die("Could not allocate memory for params->param_complement\n");
    }
    else {
      params->num_mutation_parameters[a] = 96;
      params->num_total_mutation_parameters[a] = 128;
      params->num_mutation_set[a] = 2;
    }


    (params->mutation_type[a] = (int*)malloc(sizeof(int) * params->num_mutation_parameters[a])) || die("Could not allocate memory for params->mutation_type\n");
    (params->num_mutation_parameters_per_type[a] = (int*)malloc(sizeof(int) * params->num_mutation_types)) || die("Could not allocate memory for params->num_mutation_parameters_per_type\n");
    (params->num_total_parameters_per_type[a] = (int*)malloc(sizeof(int) * (params->num_total_mutation_parameters[a] - params->num_mutation_parameters[a]))) || die("Could not allocate memory for params->num_total_parameters_per_type\n");
    (params->num_mutation_types_per_total_parameters[a] = (int*)malloc(sizeof(int) * (params->num_total_mutation_parameters[a] - params->num_mutation_parameters[a]))) || die("Could not allocate memory for params->num_mutation_types_per_total_parameters[%d]\n", a);
    (params->mutation_types_per_total_parameters[a] = (int**)malloc(sizeof(int*) * (params->num_total_mutation_parameters[a] - params->num_mutation_parameters[a]))) || die("Could not allocate memory for params->mutation_types_per_total_parameters[%d]\n", a);
    for(i = 0; i < params->num_total_mutation_parameters[a] - params->num_mutation_parameters[a]; i++)
      (params->mutation_types_per_total_parameters[a][i] = (int*)malloc(sizeof(int) * 2)) || die("Could not allocate memory for params->mutation_types_per_total_parameters[%d][%d]\n", a, i);

    (params->param_enum[a] = (int***)malloc(sizeof(int**) * params->num_total_mutation_parameters[a])) || die("Could not allocate memory for params->param_enum[%d]\n", a);
    for(i = 0; i < params->num_total_mutation_parameters[a]; i++) {
      (params->param_enum[a][i] = (int**)malloc(sizeof(int*) * params->num_mutation_set[a])) || die("Could not allocate memory for params->param_enum[%d]\n", i);
      for(j = 0; j < params->num_mutation_set[a]; j++)
	(params->param_enum[a][i][j] = (int*)malloc(sizeof(int) * 4)) || die("Could not allocate memory for params->param_enum[%d][%d][%d]\n", a, i, j);
    }
    (params->param_enum_sum[a] = (int**)malloc(sizeof(int*) * (params->num_total_mutation_parameters[a] - params->num_mutation_parameters[a]))) || die("Could not allocate memory for params->param_enum_sum[%d]\n", a);
    for(i = 0; i < params->num_total_mutation_parameters[a] - params->num_mutation_parameters[a]; i++)
      (params->param_enum_sum[a][i] = (int*)malloc(sizeof(int) * 3)) || die("Could not allocate memory for params->param_enum_sum[%d][%d]\n", a, i);
    (params->param_enum_addend[a] = (int*)malloc(sizeof(int) * params->num_mutation_parameters[a])) || die("Could not allocate memory for params->param_enum_addend[%d]\n", a);
    
    
    for(i = 0; i < 256; i++)
      params->trinuc_parameters[a][i] = -1;
    
    tmp = params->num_mutation_set[a] - 1;

    count = 0;
    for(k = 0; k < NUM_NUC; k++)
      for(l = 0; l < NUM_NUC; l++)
	for(i = 0; i < NUM_NUC; i++)
	  for(j = 0; j < NUM_NUC; j++)
	    if(l != k && params->trinuc_parameters[a][64*i + 16*k + 4*j + l] == -1) {
	      params->trinuc_parameters[a][64*i + 16*k + 4*j + l] = count;
	      params->param_enum[a][count][0][0] = i;
	      params->param_enum[a][count][0][1] = k;
	      params->param_enum[a][count][0][2] = j;
	      params->param_enum[a][count][0][3] = l;
	      params->mutation_type[a][count] = 0;
	      if(a) {
		params->complement_trinuc_parameters[64 * complement[j] + 16 * complement[k] + 4 * complement[i] + complement[l]] = count;
		params->param_complement[count] = count + 1;
		params->param_complement[count+1] = count;
		count++;
		params->complement_trinuc_parameters[64*i + 16*k + 4*j + l] = count;
	      }
	      params->trinuc_parameters[a][64 * complement[j] + 16 * complement[k] + 4 * complement[i] + complement[l]] = count;
	      params->param_enum[a][count][tmp][0] = complement[j];
	      params->param_enum[a][count][tmp][1] = complement[k];
	      params->param_enum[a][count][tmp][2] = complement[i];
	      params->param_enum[a][count][tmp][3] = complement[l];
	      count++;
	    }

    for(i = 0; i < NUM_NUC; i++)
      for(k = 0; k < NUM_NUC; k++)
	for(j = 0; j < NUM_NUC; j++)
	  if(params->trinuc_parameters[a][64*i + 17*k + 4*j] == -1) {
	    params->trinuc_parameters[a][64*i + 17*k + 4*j] = count;
	    params->param_enum[a][count][0][0] = i; 
	    params->param_enum[a][count][0][1] = k;
	    params->param_enum[a][count][0][2] = j;
	    params->param_enum[a][count][0][3] = k;
	    if(a) {
	      params->complement_trinuc_parameters[64*complement[j] + 17*complement[k] + 4*complement[i]] = count;
	      params->param_complement[count] = count + 1;
	      params->param_complement[count+1] = count;
	      count++;
	      params->complement_trinuc_parameters[64*i + 17*k + 4*j] = count;
	    }
	    params->trinuc_parameters[a][64*complement[j] + 17*complement[k] + 4*complement[i]] = count;
	    params->param_enum[a][count][tmp][0] = complement[j];
	    params->param_enum[a][count][tmp][1] = complement[k];
	    params->param_enum[a][count][tmp][2] = complement[i];
	    params->param_enum[a][count][tmp][3] = complement[k];
	    count++;
	  }
    count == params->num_total_mutation_parameters[a] || die("miscounted parameters: %d\n", count);
    
    for(i = params->num_mutation_parameters[a]; i < params->num_total_mutation_parameters[a]; i++)
      for(j = 0; j < 3; j++) {
	params->param_enum_sum[a][i - params->num_mutation_parameters[a]][j] = params->trinuc_parameters[a][64*params->param_enum[a][i][0][0] + 16*params->param_enum[a][i][0][1] +4*params->param_enum[a][i][0][2] + ((j < params->param_enum[a][i][0][3]) ? j : (j+1))];
      }
    for(i = 0; i < params->num_mutation_parameters[a]; i++) {
      params->param_enum_addend[a][i] = params->trinuc_parameters[a][64 * params->param_enum[a][i][0][0] + 16*params->param_enum[a][i][0][1] + 4*params->param_enum[a][i][0][2] + params->param_enum[a][i][0][1]];
    }

    if(params->num_mutation_types == 1)
      for(i = 0; i < params->num_mutation_parameters[a]; i++)
	params->mutation_type[a][i] = 0;
    else if(!a) {
      (in = fopen(mutation_type_file, "r")) || die("Could not open mutation_type file: %s\n", mutation_type_file);
      for(i = 0; i < params->num_mutation_parameters[a]; i++) {
	fgets(line, MAX_LINE, in) || die("Error in %s reading line: %s\n", mutation_type_file, line);
	sscanf(line, "%c%c%c->%c = %c%c%c->%c = %d", c, c+1, c+2, c+3, c+4, c+5, c+6, c+7, &value) == 9 || die("Error in %s reading line: %s\n", mutation_type_file, line);
	if(NAUC[params->param_enum[a][i][0][0]] != c[0] || NAUC[params->param_enum[a][i][0][1]] != c[1] || NAUC[params->param_enum[a][i][0][2]] != c[2] || NAUC[params->param_enum[a][i][0][3]] != c[3] || NAUC[params->param_enum[a][i][1][0]] != c[4] || NAUC[params->param_enum[a][i][1][1]] != c[5] || NAUC[params->param_enum[a][i][1][2]] != c[6] || NAUC[params->param_enum[a][i][1][3]] != c[7])
	  die("Incorrect ordering of mutations in file %s on line: %s\n", mutation_type_file, line);
	value < params->num_mutation_types || die("%d mutation types specified, but read %d in file %s\n", params->num_mutation_types, value, mutation_type_file);
	params->mutation_type[a][i] = value;
      }
    }
    else {
      for(i = 0; i < params->num_mutation_parameters[a]; i++) {
	if(fgets(line, MAX_LINE, in) && sscanf(line, "%c%c%c->%c = %d", c, c+1, c+2, c+3, &value) == 5) {
	  if(NAUC[params->param_enum[a][i][0][0]] != c[0] || NAUC[params->param_enum[a][i][0][1]] != c[1] || NAUC[params->param_enum[a][i][0][2]] != c[2] || NAUC[params->param_enum[a][i][0][3]] != c[3])
	    die("Incorrect ordering of mutations in file %s on line: %s\n", mutation_type_file, line);
	  value < params->num_mutation_types || die("%d mutation types specified, but read %d in file %s\n", params->num_mutation_types, value, mutation_type_file);
	  params->mutation_type[a][i] = value;
	}
	else
	  params->mutation_type[a][i] = params->mutation_type[0][i/2];
      }
      fclose(in);
    }

    for(i = 0; i < params->num_mutation_types; i++)
      params->num_mutation_parameters_per_type[a][i] = params->num_total_parameters_per_type[a][i] = 0;
    for(i = 0; i < params->num_mutation_parameters[a]; i++)
      params->num_mutation_parameters_per_type[a][params->mutation_type[a][i]]++;
    
    for(i = 0; i < params->num_total_mutation_parameters[a] - params->num_mutation_parameters[a]; i++) {
      params->num_mutation_types_per_total_parameters[a][i] = 0;
      for(j = 0; j < 3; j++) {
	k = params->mutation_type[a][params->param_enum_sum[a][i][j]];
	for(l = 0; l < params->num_mutation_types_per_total_parameters[a][i]; l++)
	  if(params->mutation_types_per_total_parameters[a][i][l] == k)
	    break;
	if(l == params->num_mutation_types_per_total_parameters[a][i]) {
	  params->mutation_types_per_total_parameters[a][i][params->num_mutation_types_per_total_parameters[a][i]++] = k;
	  params->num_total_parameters_per_type[a][k]++;
	}
      }
    }

    (params->mutation_parameters_per_type[a] = (int**)malloc(sizeof(int*) * params->num_mutation_types)) || die("Could not allocate memory for params->mutation_parameters_per_type\n");
    (params->total_parameters_per_type[a] = (int**)malloc(sizeof(int*) * params->num_mutation_types)) || die("Could not allocate memory for params->total_parameters_per_type[%d]\n", a);
    for(i = 0; i < params->num_mutation_types; i++) {
      if(params->num_mutation_parameters_per_type[a][i])
	(params->mutation_parameters_per_type[a][i] = (int*)malloc(sizeof(int) * params->num_mutation_parameters_per_type[a][i])) || die("Could not allocate memory for params->mutation_parameters_per_type[%d]\n", i);
      if(params->num_total_parameters_per_type[a][i])
	(params->total_parameters_per_type[a][i] = (int*)malloc(sizeof(int) * params->num_total_parameters_per_type[a][i])) || die("Could not allocate memory for params->total_parameters_per_type[%d][%d]\n", a, i);
    }
    for(i = 0; i < params->num_mutation_types; i++)
      params->num_mutation_parameters_per_type[a][i] = params->num_total_parameters_per_type[a][i] = 0;
    for(i = 0; i < params->num_mutation_parameters[a]; i++)
      params->mutation_parameters_per_type[a][params->mutation_type[a][i]][params->num_mutation_parameters_per_type[a][params->mutation_type[a][i]]++] = i;  
    for(i = 0; i < params->num_total_mutation_parameters[a] - params->num_mutation_parameters[a]; i++) {
      for(j = 0; j < params->num_mutation_types_per_total_parameters[a][i]; j++)
	params->total_parameters_per_type[a][params->mutation_types_per_total_parameters[a][i][j]][params->num_total_parameters_per_type[a][params->mutation_types_per_total_parameters[a][i][j]]++] = i;
    }
  }  

  (params->mutation_type_symmetry = (int*)malloc(sizeof(int) * params->num_mutation_types)) || die("Could not allocate memory for params->mutation_type_symmetry\n");
  
  for(i = 0; i < params->num_mutation_types; i++) {
    params->mutation_type_symmetry[i] = 0;
    if(params->num_mutation_parameters_per_type[1][i]) {
      params->mutation_type_symmetry[i] = 1;
      if(params->num_mutation_parameters_per_type[0][i])
	params->mutation_type_symmetry[i] = 2;
    }
    fprintf(stderr, "mutation_type_symmetry[%d] = %d\n", i, params->mutation_type_symmetry[i]);
  }
}
  
void make_root_parameters(Params* params) {
  int i, j, k;
  int count = 0;
  
  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++)
     for(k = 0; k < NUM_NUC; k++)
	params->root_parameters[i][j][k] = -1;
  
  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++)
      for(k = 0; k < NUM_NUC - 1; k++)
	if(params->root_parameters[i][j][k] == -1) {
	  params->root_parameters[i][j][k] = count;
	  params->root_parameters_vector[16 * i + 4 * j + k] = count;
	  params->complement_root_parameters[complement[k]][complement[j]][complement[i]] = count;
	  params->complement_root_parameters_vector[16 * complement[k] + 4 * complement[j] + complement[i]] = count;
	  params->root_enum[count][0] = i;
	  params->root_enum[count][1] = j;
	  params->root_enum[count][2] = k;
	  count++;
	}
  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++)
      if(params->root_parameters[i][j][NUM_NUC - 1] == -1) {
	params->root_parameters[i][j][NUM_NUC - 1] = count;
	params->root_parameters_vector[16 * i + 4 * j + 3] = count;
	params->complement_root_parameters[complement[NUM_NUC - 1]][complement[j]][complement[i]] = count;
	params->complement_root_parameters_vector[16 * complement[3] + 4 * complement[j] + complement[i]] = count;
	params->root_enum[count][0] = i;
	params->root_enum[count][1] = j;
	params->root_enum[count][2] = NUM_NUC - 1;
	count++;
      }
  count == NUM_TOTAL_ROOT_PARAMETERS || die("miscounted root parameters: %d\n", count);

  for(i = NUM_ROOT_PARAMETERS; i < NUM_TOTAL_ROOT_PARAMETERS; i++)
    for(j = 0; j < 3; j++)
      params->root_enum_sum[i-NUM_ROOT_PARAMETERS][j] = params->root_parameters[params->root_enum[i][0]][params->root_enum[i][1]][j];
  for(i = 0; i < NUM_ROOT_PARAMETERS; i++)
    params->root_enum_addend[i] = params->root_parameters[params->root_enum[i][0]][params->root_enum[i][1]][NUM_NUC - 1];
}

void read_rate_categories(Params* params, char* filename) {
  FILE* in;
  int c, i, j, k, l, m, last;
  char name[MAX_NAME];
  int start, end, num_regions, num_branches, *branch_start, *branch_end, category;
  int direction;
  char parameters_file[MAX_NAME];
  int tmp;
  char outname[MAX_LINE];

  for(i = 0; i < params->seq_length; i++) {
    for(j = 0; j <= params->num_branches; j++) {
      params->branch_state[i][j].rate_category = 0;
      params->branch_state[i][j].direction = 0;
    }
  }
  (in = fopen(filename, "r")) || die("Could not open rate categories file: %s\n", filename);

  (params->time_category = (int*)malloc(sizeof(int) * (params->num_branches + 1))) || die("Could not allocate memory for params->time_category\n");
  for(i = 0; i < params->num_branches; i++)
    params->time_category[i] = 0;
  fscanf(in, "%d", &params->num_time_categories) == 1 || die("Error reading number of time categories in file: %s\n", filename);
  for(c = 0; c < params->num_time_categories; c++) {
    fscanf(in, "%d", &num_branches) == 1 || die("Error reading number of branch ranges in time category %d\n", c);
    if(num_branches) 
      for(i = 0; i < num_branches; i++) {
	fscanf(in, "%d-%d", &start, &end) == 2 || die("Error reading branch range %d in time category %d\n", i, c);
	start <= end && start >= 0 && end <= params->num_branches || die("Branch range %d to %d is not valid in file: %s\n", start, end, filename);
	for(j = start; j <= end; j++)
	  params->time_category[j] = c;
      }
    else
      for(i = 0; i < params->num_branches; i++)
	params->time_category[i] = c;
  }
  fscanf(in, "%d", &params->num_rate_categories) == 1 || die("Error reading number of rate categories in file: %s\n", filename);

  (params->branch_param_table = (ParamTable***)malloc(sizeof(ParamTable**) * params->num_hidden_states)) || die("Could not allocate memory for params->branch_param_table\n");
  (params->reverse_branch_param_table = (ParamTable***)malloc(sizeof(ParamTable**) * params->num_hidden_states)) || die("Could not allocate memory for params->reverse_branch_param_table\n");
  if(params->variation)
    (params->branch_param_list = (double****)malloc(sizeof(double***) * params->num_hidden_states)) || die("Could not allocate memory for params->branch_param_list\n");
  for(i = 0; i < params->num_hidden_states; i++) {
    (params->branch_param_table[i] = (ParamTable**)malloc(sizeof(ParamTable*) * params->num_rate_categories)) || die("Could not allocate memory for params->branch_param_table[%d]\n", i);
    (params->reverse_branch_param_table[i] = (ParamTable**)malloc(sizeof(ParamTable*) * params->num_rate_categories)) || die("Could not allocate memory for params->reverse_branch_param_table[%d]\n", i);
    if(params->variation)
      (params->branch_param_list[i] = (double***)malloc(sizeof(double**) * params->num_rate_categories)) || die("Could not allocate memory for params->branch_param_list[%d]\n", i);
  }
  
  (params->category_name = (char**)malloc(sizeof(char*) * params->num_rate_categories)) || die("Could not allocate memory for params->category_name\n");
  (params->category_type = (char*)malloc(sizeof(char) * params->num_rate_categories)) || die("Could not allocate memory for params->category_type\n");
  (params->parameters = (FLOAT**)malloc(sizeof(FLOAT*) * params->num_rate_categories)) || die("Could not allocate memory for params->parameters\n");
  (params->category_branch_existence = (char**)malloc(sizeof(char*) * params->num_rate_categories)) || die("Could not allocate memory for params->category_branch_existence\n");
  (params->rate_time_category = (int*)malloc(sizeof(int) * params->num_rate_categories)) || die("Could not allocate memory for params->rate_time_category\n");
  (params->mutation_out = (FILE**)malloc(sizeof(FILE*) * params->num_rate_categories)) || die("Could not allocate memory for params->mutation_out\n");
  
  for(c = 0; c < params->num_rate_categories; c++) {
    fscanf(in, "%s %d %s %d", name, &tmp, parameters_file, params->rate_time_category + c) <= 4 || die("Error reading category in file: %s\n", filename);
    params->category_type[c] = tmp;
    params->category_name[c] = strdup(name);
    params->category_type[c] == 0 || params->category_type[c] == 1 || die("Invalid category type: %d\n", params->category_type);
    params->rate_time_category[c] >= 0 && params->rate_time_category[c] < params->num_time_categories || die("Time category %d of rate category is out of range\n", params->rate_time_category[c]);
    (params->parameters[c] = (FLOAT*)malloc(sizeof(FLOAT) * params->num_mutation_parameters[params->category_type[c]])) || die("Could not allocate memory for params->parameters[%d]\n", c);
    for(i = 0; i < params->num_hidden_states; i++) {
      (params->branch_param_table[i][c] = (ParamTable*)malloc(sizeof(ParamTable) * params->num_branches)) || die("Could not allocate memory for params->branch_param_table[%d][%d]\n", i, c);
      if(params->category_type[c])
	(params->reverse_branch_param_table[i][c] = (ParamTable*)malloc(sizeof(ParamTable) * params->num_branches)) || die("Could not allocate memory for params->reverse_branch_param_table[%d][%d]\n", i, c);
      if(params->variation) {
	(params->branch_param_list[i][c] = (double**)malloc(sizeof(double*) * params->num_branches)) || die("Could not allocate memory for params->branch_param_list[%d][%d]\n", i, c);
	for(j = 0; j < params->num_branches; j++)
	  (params->branch_param_list[i][c][j] = (double*)malloc(sizeof(double) * params->num_total_mutation_parameters[params->category_type[c]])) || die("Could not allocate memory for params->branch_param_list[%d][%d][%d]\n", i, c, j);
      }
    }
    (params->category_branch_existence[c] = (char*)malloc(sizeof(char) * params->num_branches)) || die("Could not allocate memory for params->category_branch_existence[%d]\n", c);
    for(i = 0; i < params->num_branches; i++)
      params->category_branch_existence[c][i] = 0;
    read_parameters(params, params->parameters[c], params->category_type[c], parameters_file);
    sprintf(outname, "%smcmc.%s.%s.mutation", params->out_prefix, params->category_name[c], params->category_type[c] ? "asym" : "sym");
    (params->mutation_out[c] = fopen(outname, "w")) || die("Could not create file: %s\n", outname);
  }

  /*
  for(c = 0; c < params->num_rate_categories; c++)
    for(i = 0; i < params->num_mutation_parameters[params->category_type[c]]; i++)
      fprintf(stderr, "parameter[c][i] = %g\n", params->parameters[c][i]);
  */
  (branch_start = (int*)malloc(sizeof(int) * params->num_branches)) || die("Could not allocate memory for branch_start\n");
  (branch_end = (int*)malloc(sizeof(int) * params->num_branches)) || die("Could not allocate memory for branch_end\n");
  while(1) {
    tmp = fscanf(in, "%s %d %d", name, &direction, &num_branches);
    if(tmp == EOF)
      break;
    else if(tmp < 3)
      die("Error reading listing in file: %s\n", filename);
    for(i = 0; i < params->num_rate_categories; i++)
      if(!strcmp(name, params->category_name[i])) {
	category = i;
	break;
      }
    i < params->num_rate_categories || die("Unknown category name in rate categories file: %s\n", name);
    if(num_branches)
      for(i = 0; i < num_branches; i++) {
	fscanf(in, "%d-%d", branch_start + i, branch_end + i) == 2 || die("Missing branch in branch list in file: %s\n", filename);
	branch_start[i] <= branch_end[i] && branch_start[i] >= 0 && branch_end[i] <= params->num_branches || die("Invalid branch region: %d to %d\n", branch_start[i], branch_end[i]);
      }
    else {
      branch_start[0] = 0;
      branch_end[0] = params->num_branches;
      num_branches = 1;
    }

    fscanf(in, "%d", &num_regions) == 1 || die("Missing number of regions in file: %s\n", filename);
    if(num_regions)
      for(i = 0; i < num_regions; i++) {
	fscanf(in, "%d-%d", &start, &end) == 2 || die("Missing region in region list in file: %s\n", filename);
	start > 0 && end <= params->map_length || die("Invalid region: %d to %d\n", start, end);
	last = -1;
	for(j = start - 1; j < end; j++)
	  if(params->map[j] >= 0) {
	    for(m = ((last >= 0) ? last : params->map[j]); m <= params->map[j]; m++)
	      for(k = 0; k < num_branches; k++)
		for(l = branch_start[k]; l <= branch_end[k]; l++) {
		  params->branch_state[m][l].rate_category = category;
		  params->branch_state[m][l].direction = direction;
		}
	    last = m;
	  }
      }
    else {
      for(i = 0; i < params->seq_length; i++)
	for(k = 0; k < num_branches; k++)
	  for(l = branch_start[k]; l <= branch_end[k]; l++) {
	    params->branch_state[i][l].rate_category = category;
	    params->branch_state[i][l].direction = direction;
	  }
    }
  }
    
  free(branch_start);
  free(branch_end);
  fclose(in);

  for(i = 0; i < params->seq_length; i++)
    for(j = 0; j <= params->num_branches; j++)
      if(params->branch_state[i][j].present == 1) {
	params->category_branch_existence[params->branch_state[i][j].rate_category][j] = 1;
      }

  for(i = 0; i < params->num_rate_categories; i++) {
    fprintf(stderr, "category %d contains branches: ", i);
    for(j = 0; j < params->num_branches; j++)
      if(params->category_branch_existence[i][j])
	fprintf(stderr, "%d ", j);
    fprintf(stderr, "\n");
  }
  update_branch_param_table(params);
}

void read_parameters(Params* params, FLOAT* parameters, int a, char* filename) {
  int i;
  FILE* in;
  double parameter;
  char line[MAX_LINE];
  char nuc[8];
  char alpha_table[256];

  (in = fopen(filename, "r")) || die("Could not open parameters file: %s\n", filename);
  for(i = 0; i < NUM_NUC; i++)
    alpha_table[NAUC[i]] = (char)i;
  for(i = 0; i < params->num_mutation_parameters[a]; i++) {
    fgets(line, MAX_LINE, in) || die("Truncated parameters file at line %d\n", i);
    if(a) {
      sscanf(line, "%c%c%c->%c = %lf", nuc, nuc+1, nuc+2, nuc+3, &parameter) == 5 || die("Error reading line: %s", line);
      params->param_table[alpha_table[nuc[0]]][alpha_table[nuc[1]]][alpha_table[nuc[2]]][alpha_table[nuc[3]]] = parameter;
    }
    else {
      sscanf(line, "%c%c%c->%c = %c%c%c->%c = %lf", nuc, nuc+1, nuc+2, nuc+3, nuc+4, nuc+5, nuc+6, nuc+7, &parameter) == 9 || die("Error reading line: %s", line);
      params->param_table[alpha_table[nuc[0]]][alpha_table[nuc[1]]][alpha_table[nuc[2]]][alpha_table[nuc[3]]] = parameter;
      params->param_table[alpha_table[nuc[4]]][alpha_table[nuc[5]]][alpha_table[nuc[6]]][alpha_table[nuc[7]]] = parameter;
    }
    parameters[i] = parameter;
  }

  fclose(in);
}

void read_root_categories(Params* params, char* filename) {
  FILE* in;
  int c, i, j, k, m, last;
  char name[MAX_NAME];
  int start, end, num_regions, num_branches, *branch_start, *branch_end;
  char root_file[MAX_NAME];
  int tmp, category;
  char outname[MAX_LINE];

  (params->root_category = (char*)malloc(sizeof(char) * params->seq_length)) || die("Could not allocate memory for params->root_category\n");
  for(i = 0; i < params->seq_length; i++)
    params->root_category[i] = 0;
  (in = fopen(filename, "r")) || die("could not open root categories file: %s\n", filename);
  fscanf(in, "%d", &params->num_root_categories) == 1 || die("Error reading number of root categories in file: %s\n", filename);

  (params->root_category_name = (char**)malloc(sizeof(char*) * params->num_root_categories)) || die("Could not allocate memory for params->root_category_name\n");
  (params->root_transitions = (FLOAT**)malloc(sizeof(FLOAT*) * params->num_root_categories)) || die("Could not allocate memory for params->root_transitions\n");
  (params->cond_root = (RootTable*)malloc(sizeof(RootTable) * params->num_root_categories)) || die("Could not allocate memory for params->cond_root\n");
  (params->reverse_cond_root = (RootTable*)malloc(sizeof(RootTable) * params->num_root_categories)) || die("Could not allocate memory for params->reverse_cond_root\n");
  (params->root_out = (FILE**)malloc(sizeof(FILE*) * params->num_root_categories)) || die("Could not allocate memory for params->root_out\n");

  for(c = 0; c < params->num_root_categories; c++) {
    fscanf(in, "%s %s", name, root_file) == 2 || die("Error reading root category in file: %s\n", filename);
    params->root_category_name[c] = strdup(name);
    (params->root_transitions[c] = (FLOAT*)malloc(sizeof(FLOAT) * NUM_ROOT_PARAMETERS)) || die("Could not allocate memory for params->root_transitions[%d]\n", i);
    read_root_parameters(params, params->root_transitions[c], root_file);
    sprintf(outname, "%smcmc.%s.root", params->out_prefix, params->root_category_name[c]);
    (params->root_out[c] = fopen(outname, "w")) || die("Could not create file: %s\n", outname);
  }
  update_root_table(params);

  (branch_start = (int*)malloc(sizeof(int) * params->num_branches)) || die("Could not allocate memory for branch_start\n");
  (branch_end = (int*)malloc(sizeof(int) * params->num_branches)) || die("Could not allocate memory for branch_end\n");
  while(1) {
    tmp = fscanf(in, "%s %d", name, &num_branches);
    if(tmp == EOF)
      break;
    else if(tmp < 2)
      die("Error reading listing in file: %s\n", filename);
    for(i = 0; i < params->num_root_categories; i++) 
      if(!strcmp(name, params->root_category_name[i])) {
	category = i;
	break;
      }
    i < params->num_root_categories || die("Unknown category name in root categories file: %s\n", name);
    if(num_branches)
      for(i = 0; i < num_branches; i++) {
	fscanf(in, "%d-%d", branch_start + i, branch_end + i) == 2 || die("Missing branch in branch list in file: %s\n", filename);
	branch_start[i] <= branch_end[i] && branch_start[i] >= 0 && branch_end[i] <= params->num_branches || die("Invalid branch region: %d to %d\n", branch_start[i], branch_end[i]);
      }
    else {
      branch_start[0] = 0;
      branch_end[0] = params->num_branches;
      num_branches = 1;
    }

    fscanf(in, "%d", &num_regions) == 1 || die("Could not read number of regions in file: %s\n", filename);
    if(num_regions)
      for(i = 0; i < num_regions; i++) {
	fscanf(in, "%d-%d", &start, &end) == 2 || die("Could not read region in list in file: %s\n", filename);
	start > 0 && end <= params->map_length || die("Invalid region: %d to %d\n", start, end);
	last = -1;
	for(j = start - 1; j < end; j++)
	  if(params->map[j] >= 0) {
	    for(m = ((last >= 0) ? last : params->map[j]); m <= params->map[j]; m++)
	      for(k = 0; k < num_branches; k++)
		if(params->configuration_map[params->root_position[m]].node->num >= branch_start[k] && params->configuration_map[params->root_position[m]].node->num <= branch_end[k])
		  params->root_category[m] = category;
	    last = m;
	  }
      }
    else 
      for(j = 0; j < params->seq_length; j++)
	for(k = 0; k < num_branches; k++)
	  if(params->configuration_map[params->root_position[j]].node->num >= branch_start[k] && params->configuration_map[params->root_position[j]].node->num <= branch_end[k])
	    params->root_category[j] = category;
  }
  
  free(branch_start);
  free(branch_end);
  fclose(in);
}
  
void read_root_parameters(Params* params, FLOAT* parameters, char* filename) {
  FILE* in;
  double parameter;
  char line[MAX_LINE];
  char nuc[6];
  char alpha_table[256];
  int i, j, k;
  double root[NUM_NUC][NUM_NUC][NUM_NUC];

  (in = fopen(filename, "r")) || die("Could not open root parameters file: %s\n", filename);
  for(i = 0; i < NUM_NUC; i++)
    alpha_table[NAUC[i]] = (char)i;
  for(i = 0; i < NUM_TOTAL_ROOT_PARAMETERS; i++) {
    fgets(line, MAX_LINE, in) || die("Truncated root parameters file at line %d\n", i);
    sscanf(line, "%c%c%c = %lf", nuc, nuc+1, nuc+2, &parameter) == 4 || die("Error reading line: %s", line);
    root[alpha_table[nuc[0]]][alpha_table[nuc[1]]][alpha_table[nuc[2]]] = parameter;
  }
  fclose(in);

  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++) {
      root[i][j][NUM_NUC - 1] = 1.0;
      for(k = 0; k < NUM_NUC - 1; k++) {
	root[i][j][NUM_NUC - 1] -= root[i][j][k];
	parameters[params->root_parameters[i][j][k]] = root[i][j][k];
      }
    }
}

void read_sequence(Params* params, char* seqfile) {
  FILE* in;
  int i, j, ch = EOF;
  signed char alpha_table[256];

  (in = fopen(seqfile, "r")) || die("Could not open seqfile: %s\n", seqfile);
  fscanf(in, "%d%d\n", &params->n_seq, &params->seq_length) == 2 || die("Error in seqfile: %s\n", seqfile);
  (params->sequence = (Sequence*)malloc(sizeof(Sequence) * params->n_seq)) || die("Could not allocate memory for params->sequence\n");

  for(i = 0; i < 256; i++)
    alpha_table[i] = -1;
  for(i = 0; i < 6; i++)
    alpha_table[NAUC[i]] = (char)i;
  for(i = 0; i < params->n_seq; i++) {
    fscanf(in, "%s.63", params->sequence[i].name) == 1 || die("Error in seqfile\n");
    (params->sequence[i].seq = (char*)malloc(sizeof(char) * params->seq_length)) || die("Could not allocate memory for params->sequence[%d].seq\n", i);
    for(j = 0; j < params->seq_length;) {
      if((ch = getc(in)) != EOF && alpha_table[ch] >= 0) {
	params->sequence[i].seq[j] = alpha_table[ch];
	j++;
      }
      else if(ch == '\n' || ch == EOF)
	die("Early sequence termination in sequence: %s, %d expected, %d found\n", params->sequence[i].name, params->seq_length, j + 1);
    }
    for(; ch != '\n' && ch != EOF; ch = getc(in));
  }
  
  fclose(in);
}

read_variation_file(Params* params, char* filename) {
  int i, j;
  FILE* in;
  int ch;
  double sum;

  (in = fopen(filename, "r")) || die("Could not open variation ile: %s\n", filename);

  fscanf(in, "%d", &params->num_hidden_states) == 1 || die("Error reading number of hidden states in file: %s\n", filename);

  (params->rate_multiplier = (double*)malloc(sizeof(double) * params->num_hidden_states)) || die("Could not allocate memory for params->rate_multiplier\n");
  (params->transition = (double**)malloc(sizeof(double*) * params->num_hidden_states)) || die("Could not allocate memory for params->transition\n");
  for(i = 0; i < params->num_hidden_states; i++)
    (params->transition[i] = (double*)malloc(sizeof(double) * params->num_hidden_states)) || die("Could not allocate memory for params->transition[%d]\n", i);

  (params->fix_rate_multiplier = (int*)malloc(sizeof(int) * params->num_hidden_states)) || die("Could not allocate memory for params->fix_rate_multiplier\n");
  (params->fix_transition = (int**)malloc(sizeof(int*) * params->num_hidden_states)) || die("Could not allocate memory for params->fix_transition\n");
  for(i = 0; i < params->num_hidden_states; i++)
    (params->fix_transition[i] = (int*)malloc(sizeof(int) * params->num_hidden_states)) || die("Could not allocate memory for params->fix_transition[%d]\n", i);
  
  (params->last_unfixed_transition = (int*)malloc(sizeof(int) * params->num_hidden_states)) || die("Could not allocate memory for params->last_unfixed_transition\n");

  for(i = 0; i < params->num_hidden_states; i++) {
    fscanf(in, "%lg", params->rate_multiplier + i);
    ch = getc(in);
    if(ch == '*')
      params->fix_rate_multiplier[i] = 1;
    else {
      params->fix_rate_multiplier[i] = 0;
      ungetc(ch, in);
    }
  }
  for(i = 0; i < params->num_hidden_states; i++) {
    params->last_unfixed_transition[i] = 0;
    sum = 0.0;
    for(j = 0; j < params->num_hidden_states; j++) {
      fscanf(in, "%lg", params->transition[i] + j);
      sum += params->transition[i][j];
      ch = getc(in);
      if(ch == '*')
	params->fix_transition[i][j] = 1;
      else {
	params->fix_transition[i][j] = 0;
	params->last_unfixed_transition[i] = j + 1;
	ungetc(ch, in);
      }
    }
    for(j = 0; j < params->num_hidden_states; j++)
      params->transition[i][j] /= sum;
  }
}

void markov(Params* params) {
  int trinuc[NUM_NUC][NUM_NUC][NUM_NUC];
  int count;
  int i, j, k;

  count = 0;
  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++)
      for(k = 0; k < NUM_NUC; k++)
	trinuc[i][j][k] = 0;

  for(i = 0; i < params->n_seq; i++)
    for(j = 2; j < params->seq_length; j++)
      if(params->sequence[i].seq[j-2] < NUM_NUC && params->sequence[i].seq[j-1] < NUM_NUC && params->sequence[i].seq[j] < NUM_NUC) {
        trinuc[params->sequence[i].seq[j-2]][params->sequence[i].seq[j-1]][params->sequence[i].seq[j]]++;
	count++;
      }
  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++)
      for(k = 0; k < NUM_NUC; k++) {
	params->trinuc_freqs[1][i][j][k] = (double)trinuc[i][j][k] / count;
	params->trinuc_freqs[0][i][j][k] = (double)(trinuc[i][j][k] + trinuc[complement[k]][complement[j]][complement[i]]) / (count);
      }
}

void read_tree(Params* params, char* treefile) {
  FILE* in;

  (in = fopen(treefile, "r")) || die("Could not open treefile: %s\n", treefile);
  params->num_branches = 0;
  params->root = read_node(params, in, NULL);
  fclose(in);
  params->root->num = params->num_branches;
}

Node* read_node(Params* params, FILE* in, Node* parent) {
  Node* node;
  int i, ch;
  char name[MAX_NAME];
  double branch_length;

  (node = (Node*)malloc(sizeof(Node))) || die("Could not allocate memory for node\n");
  node->parent = parent;
  node->n_children = 0;

  while((ch = getc(in)) == '\n');
  if(ch == '(') {
    while(1) {
      while(isspace(ch = getc(in)));
      if(ch == ')')
	break;
      node->n_children < MAX_CHILDREN || die("Node has too many children\n");
      if(node->n_children)
	ch == ',' || die("Missing comma in treefile\n");
      else
	ungetc(ch, in);
      node->children[node->n_children++] = read_node(params, in, node);
      node->seq_num = -1;
    }
  }
  else {
    ungetc(ch, in);
    fscanf(in, "%[^\n:].63", name) == 1 || die("Missing name in treefile\n");
    for(i = 0; i < params->n_seq; i++)
      if(!strcmp(params->sequence[i].name, name))
	break;
    if(i == params->n_seq)
      die("Name %s in tree file but not found in sequence file\n", name);
    node->seq_num = i;
  }
  if(parent) {
    while(isspace(ch = getc(in)));
    ch == ':' || die("Missing colon in treefile after %s\n", name);
    while(isspace(ch = getc(in)));
    ungetc(ch, in);
    fscanf(in, "%lf", &branch_length) == 1 || die("Missing time to parent\n"); 
    node->time_to_parent = (int)(branch_length + 2.0);
    params->branch_length[params->num_branches] = node->time_to_parent;
    (params->branch_ratio[params->num_branches] = (double*)malloc(sizeof(double) * params->num_mutation_types)) || die("Could not allocate memory for params->branch_ratio[%d]\n", params->num_branches);
    for(i = 0; i < params->num_mutation_types; i++)
      params->branch_ratio[params->num_branches][i] = branch_length / node->time_to_parent;
    node->num = params->num_branches++;
  }
  else {
    node->time_to_parent = 1;
    while(isspace(ch = getc(in)));
    ch == ';' || die("Missing semicolon at end of treefile: %c\n", ch);
  }
  return node;
}









