/*
  mutation_sum program from the AMBIORE package.

  (C) Copyright 2004, Dick Hwang and University of Washington.

  dhwang@u.washington.edu
  http://www.phrap.org

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <ctype.h>
#include "utility.h"

#define USAGE "USAGE: mutation_sum [-asymmetry] [-asymdiff] [-nointerleave] [-scale scale] [-log] num_samples mcmc_output\n"

#define MAX_LINE 8192
#define MAX_NAME 64
#define MAX_BRANCHES 64

#define NUM_NUC 4
#define NUM_ROOT_PARAMETERS 48
#define NUM_TOTAL_ROOT_PARAMETERS 64
#define NUM_ROOT_BASE 16

#define MAX_CHILDREN 16

typedef struct NodeS {
  struct NodeS *children[MAX_CHILDREN];
  char name[MAX_NAME];
  int n_children;
  double time_to_parent;
  int branch_num;
} Node;

typedef struct {
  int n_samples;
  int num_branches;
  Node* root;
  int num_mutation_parameters;
  int num_total_mutation_parameters;
  int num_mutation_set;
  int asymmetry;
  int difference;
  double* mutation_rate;
  double* mutation_variance;
  double asymmetry_difference[96];
  double asymmetry_variance[96];
  double asymmetry_covariance[96];
  double root_transition[NUM_TOTAL_ROOT_PARAMETERS];
  double root_variance[NUM_TOTAL_ROOT_PARAMETERS];
  double branch_length[MAX_BRANCHES];
  double branch_variance[MAX_BRANCHES];
  int trinuc_parameters[NUM_NUC][NUM_NUC][NUM_NUC][NUM_NUC];
  int*** param_enum;
  int** param_enum_sum;
  int* param_enum_addend;
  int root_parameters[NUM_NUC][NUM_NUC][NUM_NUC];
  int root_enum[NUM_TOTAL_ROOT_PARAMETERS][3];
  int root_enum_sum[NUM_ROOT_BASE][NUM_NUC];
  int root_enum_addend[NUM_ROOT_PARAMETERS];
  int num_parameters;
  double scale;
  float** estimates;
  double* variance;
  int* window;
  int interleave;
  int log;
} Params;

char NAUC[7] = "TCAGN-";
int complement[4] = {2,3,0,1};

void read_params(int argc, char* argv[], Params* params);
void make_trinuc_parameters(Params* params);
void make_root_parameters(Params* params);
void print_parameters(Params* params);
void read_tree(Params* params, char* treefile);
Node* read_node(Params* params, FILE* in, int is_root);
void print_tree(Params* params);
void print_node(Params* params, Node* node, int is_root);
void read_bayes_output(Params* params, char* filename);
double squareroot(double x, double scale);
void calc_autocovariance(Params* params);


int main(int argc, char* argv[]) {
  Params* params;
  int i;

  (params = (Params*)malloc(sizeof(Params))) || die("Could not allocate memory for params\n");
  read_params(argc, argv, params);
  fprintf(stderr, "Calculating autocovariances...\n");
  calc_autocovariance(params);

  print_parameters(params);

  return 0;
}

double squareroot(double x, double scale) {
  if(fabs(x) < 10e-10 * scale * scale)
    return 0.0;
  else
    return sqrt(x);
}

void print_parameters(Params* params) {
  int i, tmp;
  double sum;

  sum = 0.0;
  for(i = 0; i < params->num_mutation_parameters; i++) {
    if(params->asymmetry) {
      if(params->interleave)
	printf("%c%c%c->%c\t%9.4g\t%9.4g",
	       NAUC[params->param_enum[i][0][0]], NAUC[params->param_enum[i][0][1]],
	       NAUC[params->param_enum[i][0][2]], NAUC[params->param_enum[i][0][3]],
	       params->mutation_rate[i] * params->scale, 
	       1.96 * squareroot(params->mutation_variance[i] + params->variance[i], params->mutation_rate[i]) * params->scale);
      else {
	if(2 * i < params->num_mutation_parameters)
	  tmp = 2 * i;
	else
	  tmp = 2 * i - params->num_mutation_parameters + 1;
	printf("%c%c%c->%c\t%9.4g\t%9.4g",
	       NAUC[params->param_enum[tmp][0][0]], NAUC[params->param_enum[tmp][0][1]],
	       NAUC[params->param_enum[tmp][0][2]], NAUC[params->param_enum[tmp][0][3]],
	       params->mutation_rate[tmp] * params->scale, 
	       1.96 * squareroot(params->mutation_variance[tmp] + params->variance[tmp], params->mutation_rate[tmp]) * params->scale);
      }
      if((i % 2) == 0 && params->interleave)
	printf("  diff = %9.4g +/- %9.4g (%6.3g) %c, cor = %9.4g, diff = %9.4g +/- %9.4g (%9.4g)\n", params->asymmetry_difference[i/2], 1.96 * squareroot(params->asymmetry_variance[i/2] + params->variance[params->num_mutation_parameters + i/2], params->asymmetry_difference[i/2]), params->asymmetry_difference[i/2] / squareroot(params->asymmetry_variance[i/2], params->asymmetry_difference[i/2]), fabs(params->asymmetry_difference[i/2]) > 1.96 * squareroot(params->asymmetry_variance[i/2], params->asymmetry_difference[i/2]) ? '*' : ' ', 
	       params->asymmetry_covariance[i/2] / sqrt(params->mutation_variance[i] * params->mutation_variance[i+1]), 
	       params->mutation_rate[i+1] / params->mutation_rate[i] - 1.0, 
	       1.96*sqrt(pow(params->mutation_rate[i+1] / params->mutation_rate[i], 2.0) * ((params->mutation_variance[i+1] + params->variance[i+1]) / pow(params->mutation_rate[i+1], 2.0) + (params->mutation_variance[i] + params->variance[i]) / pow(params->mutation_rate[i], 2.0) - 2.0 * params->asymmetry_covariance[i/2] / (params->mutation_rate[i+1] * params->mutation_rate[i]))), 
	       1.96*sqrt(pow(params->mutation_rate[i+1] / params->mutation_rate[i], 2.0) * (params->mutation_variance[i+1] / pow(params->mutation_rate[i+1], 2.0) + params->mutation_variance[i] / pow(params->mutation_rate[i], 2.0))));
      else
	printf("\n");
    }
    else {
      printf("%c%c%c->%c = %c%c%c->%c\t%15.9g\t%15.9g\n",
	     NAUC[params->param_enum[i][0][0]], NAUC[params->param_enum[i][0][1]],
	     NAUC[params->param_enum[i][0][2]], NAUC[params->param_enum[i][0][3]],
	     NAUC[params->param_enum[i][1][0]], NAUC[params->param_enum[i][1][1]],
	     NAUC[params->param_enum[i][1][2]], NAUC[params->param_enum[i][1][3]],
	     params->mutation_rate[i] * params->scale, 1.96 * squareroot(params->mutation_variance[i] + params->variance[i], params->mutation_rate[i]) * params->scale);
    }
    sum += params->variance[i] / params->mutation_variance[i];
  }
  /*
     params->estimates[params->num_parameters - 1][s] = 0.0;
    for(i = 0; i < params->num_branches; i++) {
      params->branch_length[i] += value;
      params->branch_length[params->num_branches] += value;
      branch_squared[i] += value * value;
      params->estimates[params->num_mutation_parameters + i][s] = value;
      params->estimates[params->num_parameters - 1][s] += value;
    }
    branch_squared[params->num_branches] += params->estimates[params->num_parameters - 1][s] * params->estimates[params->num_parameters - 1][s];

    
    for(i = 0; i < NUM_ROOT_BASE; i++)
      sum[i] = 1.0;
    for(i = 0; i < NUM_ROOT_PARAMETERS; i++) {
      params->root_transition[i] += value;
      root_squared[i] += value * value;
      sum[params->root_enum_addend[i] - NUM_ROOT_PARAMETERS] -= value;
      params->estimates[params->num_mutation_parameters + params->num_branches + i][s] = value;
    }
    for(i = 0; i < NUM_ROOT_BASE; i++) {
      params->root_transition[i + NUM_ROOT_PARAMETERS] += sum[i];
      root_squared[i + NUM_ROOT_PARAMETERS] += sum[i] * sum[i];
      params->estimates[params->num_mutation_parameters + params->num_branches + NUM_ROOT_PARAMETERS + i][s] = sum[i];
    }
    }
      */
  /*
  printf("variance ratio is  %g\n", sum / 96);
  */
}

void read_params(int argc, char* argv[], Params* params) {
  int i;

  params->asymmetry = 0;
  params->difference = 0;
  params->scale = 1.0;
  params->interleave = 1;
  params->log = 0;


  for(i = 1; i < argc && argv[i][0] == '-'; i++) {
    if(!strcmp(argv[i] + 1, "asymmetry"))
      params->asymmetry = 1;
    else if(!strcmp(argv[i] + 1, "scale"))
      params->scale = atof(argv[++i]);
    else if(!strcmp(argv[i] + 1, "asymdiff")) {
      params->difference = 1;
      params->asymmetry = 1;
    }
    else if(!strcmp(argv[i] + 1, "nointerleave")) {
      params->interleave = 0;
      params->asymmetry = 1;
    }
    else if(!strcmp(argv[i] + 1, "log"))
      params->log = 1;
    else
      die(USAGE);
  }
	
  argc == i + 2 || die(USAGE);
  params->n_samples = atoi(argv[i++]);
  make_trinuc_parameters(params);
  make_root_parameters(params);
  params->num_parameters = params->num_mutation_parameters + (params->asymmetry ? 96 : 0);
  read_bayes_output(params, argv[i]);
}

void make_trinuc_parameters(Params* params) {
  int i, j, k, l;
  int count = 0;
  int tmp;

  if(params->asymmetry) {
    params->num_mutation_parameters = 192;
    params->num_total_mutation_parameters = 256;
    params->num_mutation_set = 1;
  }
  else {
    params->num_mutation_parameters = 96;
    params->num_total_mutation_parameters = 128;
    params->num_mutation_set = 2;
  }

  (params->param_enum = (int***)malloc(sizeof(int**) * params->num_total_mutation_parameters)) || die("Could not allocate memory for params->param_enum\n");
  for(i = 0; i < params->num_total_mutation_parameters; i++) {
    (params->param_enum[i] = (int**)malloc(sizeof(int*) * params->num_mutation_set)) || die("Could not allocate memory for params->param_enum[%d]\n", i);
    for(j = 0; j < params->num_mutation_set; j++)
      (params->param_enum[i][j] = (int*)malloc(sizeof(int) * 4)) || die("Could not allocate memory for params->param_enum[%d][%d]\n", i, j);
  }
  (params->param_enum_sum = (int**)malloc(sizeof(int*) * (params->num_total_mutation_parameters - params->num_mutation_parameters))) || die("Could not allocate memory for params->param_enum_sum\n");
  for(i = 0; i < params->num_total_mutation_parameters - params->num_mutation_parameters; i++)
    (params->param_enum_sum[i] = (int*)malloc(sizeof(int) * 3)) || die("Could not allocate memory for params->param_enum_sum[%d]\n", i);
  (params->param_enum_addend = (int*)malloc(sizeof(int) * params->num_mutation_parameters)) || die("Could not allocate memory for params->param_enum_addend\n");

  for(i = 0; i < NUM_NUC; i++)
    for(k = 0; k < NUM_NUC; k++)
      for(j = 0; j < NUM_NUC; j++)
	for(l = 0; l < NUM_NUC; l++)
	  params->trinuc_parameters[i][k][j][l] = -1;
  
  tmp = params->num_mutation_set - 1;

  for(k = 0; k < NUM_NUC; k++)
    for(l = 0; l < NUM_NUC; l++)
      for(i = 0; i < NUM_NUC; i++)
	for(j = 0; j < NUM_NUC; j++)
	  if(l != k && params->trinuc_parameters[i][k][j][l] == -1) {
	    params->trinuc_parameters[i][k][j][l] = count;
	    params->param_enum[count][0][0] = i;
	    params->param_enum[count][0][1] = k;
	    params->param_enum[count][0][2] = j;
	    params->param_enum[count][0][3] = l;
	    if(params->asymmetry)
	      count++;
	    params->trinuc_parameters[complement[j]][complement[k]][complement[i]][complement[l]] = count;
	    params->param_enum[count][tmp][0] = complement[j];
	    params->param_enum[count][tmp][1] = complement[k];
	    params->param_enum[count][tmp][2] = complement[i];
	    params->param_enum[count][tmp][3] = complement[l];
	    count++;
	  }
  for(i = 0; i < NUM_NUC; i++)
    for(k = 0; k < NUM_NUC; k++)
      for(j = 0; j < NUM_NUC; j++)
	if(params->trinuc_parameters[i][k][j][k] == -1) {
	  params->trinuc_parameters[i][k][j][k] = count;
	  params->param_enum[count][0][0] = i;
	  params->param_enum[count][0][1] = k;
	  params->param_enum[count][0][2] = j;
	  params->param_enum[count][0][3] = k;
	  if(!params->asymmetry) {
	    params->trinuc_parameters[complement[j]][complement[k]][complement[i]][complement[k]] = count;
	    params->param_enum[count][1][0] = complement[j];
	    params->param_enum[count][1][1] = complement[k];
	    params->param_enum[count][1][2] = complement[i];
	    params->param_enum[count][1][3] = complement[k];
	  }
	  count++;
	}
  count == params->num_total_mutation_parameters || die("miscounted parameters: %d\n", count);

  for(i = params->num_mutation_parameters; i < params->num_total_mutation_parameters; i++)
    for(j = 0; j < 3; j++)
      params->param_enum_sum[i - params->num_mutation_parameters][j] = params->trinuc_parameters[params->param_enum[i][0][0]][params->param_enum[i][0][1]][params->param_enum[i][0][2]][(j < params->param_enum[i][0][3]) ? j : (j+1)];
  for(i = 0; i < params->num_mutation_parameters; i++)
    params->param_enum_addend[i] = params->trinuc_parameters[params->param_enum[i][0][0]][params->param_enum[i][0][1]][params->param_enum[i][0][2]][params->param_enum[i][0][1]];
}

void make_root_parameters(Params* params) {
  int i, j, k;
  int count = 0;

  for(i = 0; i < NUM_NUC; i++)
    for(k = 0; k < NUM_NUC; k++)
      for(j = 0; j < NUM_NUC; j++)
	params->root_parameters[i][j][k] = -1;
  
  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++)
      for(k = 0; k < NUM_NUC - 1; k++)
	if(params->root_parameters[i][j][k] == -1) {
	  params->root_parameters[i][j][k] = count;
	  params->root_enum[count][0] = i;
	  params->root_enum[count][1] = j;
	  params->root_enum[count][2] = k;
	  count++;
	}
  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++)
      if(params->root_parameters[i][j][NUM_NUC - 1] == -1) {
	params->root_parameters[i][j][NUM_NUC - 1] = count;
	params->root_enum[count][0] = i;
	params->root_enum[count][1] = j;
	params->root_enum[count][2] = NUM_NUC - 1;
	count++;
      }
  count == NUM_TOTAL_ROOT_PARAMETERS || die("miscounted root parameters: %d\n", count);

  for(i = NUM_ROOT_PARAMETERS; i < NUM_TOTAL_ROOT_PARAMETERS; i++)
    for(j = 0; j < 3; j++)
      params->root_enum_sum[i-NUM_ROOT_PARAMETERS][j] = params->root_parameters[params->root_enum[i][0]][params->root_enum[i][1]][j];
  for(i = 0; i < NUM_ROOT_PARAMETERS; i++)
    params->root_enum_addend[i] = params->root_parameters[params->root_enum[i][0]][params->root_enum[i][1]][NUM_NUC - 1];
}


void calc_autocovariance(Params* params) {
  int t, i, s, tmp;
  double value, min;
  float *e, *et, *et1;
  
  (params->window = (int*)malloc(sizeof(int) * params->num_parameters)) || die("Could not allocate memory for params->window\n");
  (params->variance = (double*)malloc(sizeof(double) * params->num_parameters)) || die("Could not allocate memory for params->variance\n");
  
  for(i = 0; i < params->num_parameters; i++) {
    value = 0.0;
    e = params->estimates[i];
    for(s = 0; s < params->n_samples; s++)
      value += *e * *e++;
    params->variance[i] = -value;
    for(t = 0; t < params->n_samples - 1; t += 2) {
      e = params->estimates[i];
      et = params->estimates[i] + t;
      et1 = params->estimates[i] + t + 1;
      value = 0.0;
      for(s = 0; s < params->n_samples - t - 1; s++)
	value += *e++ * (*et++ + *et1++);
      value += *e * *et;
      if(value <= 0.0) {
	params->window[i] = t - 2;
	break;
      }
      else if(value < min || !t)
	min = value;
      else
	value = min;
      params->variance[i] += 2.0 * value;
    }
    params->variance[i] /= (double)((double)params->n_samples * (double)params->n_samples);
    /*
    fprintf(stderr, "window[%d] = %d, variance = %g\n", i, params->window[i], params->variance[i]);
    */
  }
}

void read_bayes_output(Params* params, char* filename) {
  FILE* in;
  int s, i;
  double *mutation_squared, root_squared[NUM_TOTAL_ROOT_PARAMETERS], branch_squared[MAX_BRANCHES];
  double value, sum[NUM_ROOT_BASE];
  double last;

  (params->mutation_rate = (double*)malloc(sizeof(double) * params->num_mutation_parameters)) || die("Could not allocate memory for params->mutation_rate\n");
  (params->mutation_variance = (double*)malloc(sizeof(double) * params->num_mutation_parameters)) || die("Could not allocate memory for params->mutation_variance\n");
  (mutation_squared = (double*)malloc(sizeof(double) * params->num_mutation_parameters)) || die("Could not allocate memory for mutation_squared\n");
  (params->estimates = (float**)malloc(sizeof(float*) * params->num_parameters)) || die("Could not allocate memory for params->estimates\n");
  for(i = 0; i < params->num_parameters; i++)
    (params->estimates[i] = (float*)malloc(sizeof(float) * params->n_samples)) || die("Could not allocate memory for params->estimates[%d]\n", i);
  

  for(i = 0; i < params->num_mutation_parameters; i++)
    params->mutation_rate[i] = mutation_squared[i] = 0.0;
  for(i = 0; i < 96; i++)
    params->asymmetry_difference[i] = params->asymmetry_variance[i] = params->asymmetry_covariance[i] = 0.0;
  for(i = 0; i < NUM_TOTAL_ROOT_PARAMETERS; i++)
    params->root_transition[i] = root_squared[i] = 0.0;
  for(i = 0; i <= params->num_branches; i++)
    params->branch_length[i] = branch_squared[i] = 0.0;  

  (in = fopen(filename, "r")) || die("Could not open bayes output file: %s\n", filename);
  for(s = 0; s < params->n_samples; s++) {
    for(i = 0; i < params->num_mutation_parameters; i++) {
      fscanf(in, "%lg", &value) || die("Could not read mutation parameter %d on line %d\n", i, s);
      if(params->log)
	value = log(value);
      params->mutation_rate[i] += value;
      mutation_squared[i] += value * value;
      params->estimates[i][s] = value;
      if(params->asymmetry && (i % 2) == 1) {
	if(params->difference) {
	  params->asymmetry_difference[i/2] += - (last - value) / last;
	  params->asymmetry_variance[i/2] += (last - value) * (last - value) / (last * last);
	  params->estimates[params->num_mutation_parameters + i/2][s] = - (last - value) / last;
	}
	else {
	  params->asymmetry_difference[i/2] += - (last - value); /* last; */
	  params->asymmetry_variance[i/2] += (last - value) * (last - value); /* (last * last); */
	  params->estimates[params->num_mutation_parameters + i/2][s] = -(last - value);
	}
	params->asymmetry_covariance[i/2] += last * value;
      }
      else 
	last = value;
    }
  }
  for(i = 0; i < params->num_mutation_parameters; i++) {
    params->mutation_rate[i] /= s;
    params->mutation_variance[i] = mutation_squared[i] / s - params->mutation_rate[i] * params->mutation_rate[i];
    for(s = 0; s < params->n_samples; s++)
      params->estimates[i][s] -= params->mutation_rate[i];
  }
  if(params->asymmetry)
    for(i = 0; i < 96; i++) {
      params->asymmetry_difference[i] /= s;
      params->asymmetry_variance[i] = params->asymmetry_variance[i] / s - params->asymmetry_difference[i] * params->asymmetry_difference[i];
      params->asymmetry_covariance[i] = params->asymmetry_covariance[i] / s - params->mutation_rate[2*i] * params->mutation_rate[2*i + 1];
      for(s = 0; s < params->n_samples; s++)
	params->estimates[params->num_mutation_parameters + i][s] -= params->asymmetry_difference[i];
      
    }
  /*
  for(i = 0; i < NUM_TOTAL_ROOT_PARAMETERS; i++) {
    params->root_transition[i] /= s;
    params->root_variance[i] = root_squared[i] / s - params->root_transition[i] * params->root_transition[i];
    for(s = 0; s < params->n_samples; s++)
      params->estimates[params->num_mutation_parameters + params->num_branches + i][s] -= params->root_transition[i];
  }
  for(i = 0; i <= params->num_branches; i++) {
    params->branch_length[i] /= s;
    if(i == params->num_branches)
      fprintf(stderr, "branch_squared = %14.9g, branch_length = %14.9g\n", branch_squared[i], params->branch_length[i]);
    params->branch_variance[i] = branch_squared[i] / s - params->branch_length[i] * params->branch_length[i];  
    for(s = 0; s < params->n_samples; s++)
      if(i < params->num_branches)
	params->estimates[params->num_mutation_parameters + i][s] -= params->branch_length[i];
      else
	params->estimates[params->num_parameters - 1][s] -= params->branch_length[i];
  }
    */
  free(mutation_squared);
}



void read_tree(Params* params, char* treefile) {
  FILE* in;

  (in = fopen(treefile, "r")) || die("Could not open treefile: %s\n", treefile);
  params->num_branches = 0;
  params->root = read_node(params, in, 1);
  fclose(in);
}

Node* read_node(Params* params, FILE* in, int is_root) {
  Node* node;
  int ch;

  (node = (Node*)malloc(sizeof(Node))) || die("Could not allocate memory for node\n");
  node->n_children = 0;

  while(isspace(ch = getc(in)));
  if(ch == '(') {
    for(; ; node->n_children++) {
      while(isspace(ch = getc(in)));
      if(ch == ')')
	break;
      if(ch != ',')
	ungetc(ch, in);
      node->n_children < MAX_CHILDREN || die("Node has too many children\n");
      !node->n_children || ch == ',' || die("Missing comma in treefile\n");
      node->children[node->n_children] = read_node(params, in, 0);
    }
  }
  else {
    ungetc(ch, in);
    fscanf(in, "%[^\n:].63", node->name) == 1 || die("Missing name in treefile\n");
  }
  if(!is_root) {
    while(isspace(ch = getc(in)));
    ch == ':' || die("Missing colon in treefile after %s\n", node->name);
    while(isspace(ch = getc(in)));
    ungetc(ch, in);
    fscanf(in, "%lg", &node->time_to_parent) == 1 || die("Missing time to parent\n"); 
    node->branch_num = params->num_branches++;
  }
  else {
    while(isspace(ch = getc(in)));
    ch == ';' || die("Missing semicolon at end of treefile\n");
  }
  return node;
}

void print_tree(Params* params) {
  if(params->root)
    print_node(params, params->root, 1);
}

void print_node(Params* params, Node* node, int is_root) {
  int n;

  if(node->n_children > 0) {
    printf("(");
    for(n = 0; n < node->n_children; n++) {
      if(n)
	printf(",");
      print_node(params, node->children[n], 0);
    }
    printf(")");
  }
  else
    printf(node->name);
  if(!is_root)
    printf(":%g+/-%g", 3.0 * params->branch_length[node->branch_num] / params->scale, 3.0 * 1.96 * squareroot(params->branch_variance[node->branch_num] + params->variance[params->num_mutation_parameters + node->branch_num], params->branch_length[node->branch_num]) / params->scale);
  else
    printf(";\n");
}






